observatories = { 
					'STANDARD': {
						'T': 288.15, 
						'p': 101325,
						'RH': 0.0,
						'xc': 450,
						'latitude': 0,
						'altitude': 0,
						'comment': 'Standard atmospheric conditions'
						},
					'CERRO_ARMAZONES': {
						'T': 279.65, 
						'p': 71200,
						'RH': 0.22,
						'xc': 450,
						'latitude': -24.5983,
						'altitude': 3064,
						'comment': 'Default conditions for Cerro Armazones, site of the ELT.\nValues are taken from the optical model of MICADO.'
						},
					'CERRO_PARANAL': {
						'T': 285.95, 
						'p': 74300,
						'RH': 0.15,
						'xc': 450,
						'latitude': -24.6230,
						'altitude': 2635,
						'comment': 'Average conditions at Cerro Paranal, site of the VLT.\nValues are taken from https://www.eso.org/sci/facilities/paranal/astroclimate/site.html and https://academic.oup.com/mnras/article/399/2/783/1061377'
						},
					'LA_SILLA': {
						'T': 284.65, 
						'p': 77220,
						'RH': 0.44,
						'xc': 450,
						'latitude': -29.25,
						'altitude': 2400,
						'comment': 'Average conditions of the La Silla Observatory.\nValues based on https://www.eso.org/gen-fac/pubs/astclim/lasilla/diffrefr.html'
						},
					'LAS_CAMPANAS': {
						'T': 285.85, 
						'p': 76400,
						'RH': 0.0,
						'xc': 450,
						'latitude': -29.016,
						'altitude': 2380,
						'comment': 'Mean conditions at Las Campanas observatory, site of the GMT.\nValues based on https://www.gmto.org/Resources/GMT-ID-01466-Chapter_5_Site_Evaluation.pdf and http://weather.lco.cl/'
						},
					'LA_PALMA': {
						'T': 281.95, 
						'p': 77500,
						'RH': 0.36,
						'xc': 450,
						'latitude': 28.764,
						'altitude': 2396,
						'comment': 'Average conditions for Roque de los Muchachos Observatory on La Palma, Spain.\nValues taken from https://academic.oup.com/mnras/article/399/2/783/1061377'
						},
					'MAUNA_KEA': {
						'T': 274.45, 
						'p': 61400,
						'RH': 0.48,
						'xc': 450,
						'latitude': 19.826,
						'altitude': 4205,
						'comment': 'Average conditions at the Mauna Kea Observatory.\nValues taken from http://www.soest.hawaii.edu/MET/Faculty/businger/mauna_kea/daSilvaThesisFinal.pdf, Ch. 3. and https://www.eso.org/gen-fac/pubs/astclim/espas/espas_reports/ESPAS-MaunaKea.pdf'
						}
					# 'MT_GRAHAM': {
					# 	'T': 278.15, 
					# 	'p': 68500,
					# 	'RH': 0.0,
					# 	'xc': 450,
					# 	'latitude': 32.701,
					# 	'altitude': 3191,
					# 	'comment': 'Mean conditions at Mount Graham International Observatory,\
					# 				site of the LBT. Values are estimates.'
					# 	},
					# 'FRED_LAWRENCE_WHIPPLE': {
					# 	'T': 288.15, 
					# 	'p': 74300,
					# 	'RH': 0.0,
					# 	'xc': 450,
					# 	'latitude': 31.681,
					# 	'altitude': 2616,
					# 	'comment': 'Mean conditions at Fred Lawrence Whipple Observatory, site of\
					# 				the MMT. Values are estimates.'
					#	}
					}


#!/usr/bin/env Python3
import numpy as np
from scipy import integrate

'''
@ Author:		Joost van den Born 
@ Contact:		born@astron.nl
@ Description:	This script offers a complete model of the atmospheric refraction
				and dispersion. It uses the Ciddor1996 paper to calculate the 
				index of refraction for a wide range of conditions, with corresponding
				error propagation if needed.
				The actual refraction is calculated by any of the following models:
				- Plane parallel atmosphere
				- Cassini spherical homogeneous atmosphere
				- Other methods described in Corbard et al., 2019.
'''

class Observatory:
	'''
		Class used for all function to calculate the refractive index
		and the corresponding uncertainties.

		...

		Attributes
		----------
		dT 	: 	float (optional)
			Uncertainty in the atmospheric temperature in Kelvin
		dP 	: 	float (optional)
			Uncertainty in the atmospheric pressure in Pascal
		dH 	: 	float (optional)
			Uncertainty in the atmospheric relative humidity
		dCO2: 	float (optional)
			Uncertianty in the atmospheric CO2 density in parts
			per million.
		dl  : 	float (optional)
			Uncertainty in the wavelengths in micron

		Methods
		-------
		n_as(l)
			Returns the refractive index for standard atmospheric conditions.
		n_axs(l, xc)
			Same as n_as(l), but with a different CO2 density given by xc
		n_ws(l)
			Returns the refractive index of air for standard moist contions.
		f(p,T)
			Returns the enchancement factor of water vapour in air.
		svp(T)
			Returns the saturation vapour pressure in Pa.
		svp2(T)
			Returns the saturation vapour pressure in Pa. 
			From IAPWS formula, based on documentation by NIST toolbox.
		xw(p, T, RH)
			Returns the molar fraction of water vapour in moist air.
		Z(p, T, RH, xw='')
			Returns the compressibility of moist air
		rho(p, T, RH, xc, xw='')
			Returns the density of atmospheric air at the given parameters.
		n_tph(l, T, p, RH, xc)
			Returns the refractive index for any atmospheric conditions.
		dn_tph(l, T, p, RH, xc, dl=None, dT=None, dP=None, dRH=None, dCO2=None)
			Returns the uncertainty in the refractive index for any atmospheric
			conditions, given the parameters.
	'''
	def __init__(self, dT=0.2, dP=20, dH=0.02, dCO2=20, dl=0.001):
		self.dT 	= dT 	# in K
		self.dP 	= dP 	# in Pa
		self.dH 	= dH	# relative humidity between 0 and 1
		self.dCO2 	= dCO2	# in ppm
		self.dl 	= dl 	# in micron

	def n_as(self, l):
		'''
			Formula that calculate the refractive index for standard atmospheric
			conditions, i.e. 15 C, 101,325 Pa, 0% RH, 450 ppm CO2.
			It returns n_as, not (n_as - 1) * 1e8!

			Parameters
			----------
			l 	: float
				Wavelength of light in microns

			Returns
			-------
			n 	: float
				Refractive index of air at standard atmospheric conditions.

		'''
		_s = 1 / l
		_k0 = 238.0185 	# um^-2
		_k1 = 5792105  	# um^-2
		_k2 = 57.362		# um^-2
		_k3 = 167917		# um^-2
		return (_k1 / (_k0 - _s**2) + _k3 / (_k2 - _s**2)) / 1e8 + 1


	def n_axs(self, l, xc):
		'''
			Formula that calculate the refractive index for standard atmospheric
			conditions, i.e. 15 C, 101,325 Pa, 0% RH, but with xc ppm of CO2.
			It returns n_axs, not (n_axs - 1)

			Parameters
			----------
			l 	: float
				Wavelength of light in microns
			xc 	: float
				CO2 density in parts per million

			Returns
			-------
			n 	: float
				Refractive index of air
		'''
		_n_as = self.n_as(l)
		return ( (_n_as - 1) * (1 + 0.534e-6 * (xc - 450) ) ) + 1


	def n_ws(self, l):
		'''
			Formula that calculates the refractivity of air for standard moist conditions
			as defined by B&S (1939), i.e. 20 C, 1333 Pa
			Returns n_ws, not (n_ws - 1) * 1e8

			Parameters
			----------
			l 	: float
				Wavelength of light in microns

			Returns
			-------
			n 	: float
				Refractive index of standard moist air.
		'''
		_s  = 1 / l
		_cf = 1.022
		_w0 = 295.235 	# um^-2
		_w1 = 2.6422	# um^-2
		_w2 = -0.032380	# um^-4
		_w3 = 0.004028	# um^-6
		return (_cf * (_w0 + _w1 * _s**2 + _w2 * _s**4 + _w3 * _s**6) ) / 1e8 + 1


	def f(self, p, T):
		'''
			Enhancement factor of water vapour in air

			Parameters
			----------
			p 	: float
				Atmospheric pressure in Pascal
			T 	: float
				Atmospheric temperature in Kelvin

			Returns
			-------
			f 	: float
				Enhancement factor of water vapour in air
		'''
		_t = T - 273.15
		_a = 1.00062
		_b = 3.14e-8 	# Pa^-1
		_c = 5.6e-7		# deg C^-2
		return _a + _b * p + _c * _t**2


	def svp(self, T):
		'''
			Saturation vapour pressure in Pa.

			Parameters
			----------
			T 	: float
				Atmospheric temperature in Kelvin

			Returns
			-------
			svp	: float
				Saturation vapour pressure
		'''
		_A =  1.2378847e-5 # K^-2
		_B = -1.9121316e-2 # K^-2
		_C = 33.93711047
		_D = -6.3431645e3  # K^-2
		return np.exp(_A*T**2 + _B*T + _C + _D/T)


	def svp2(self, T):
		'''
			Calculation of svp from IAPWS formula, based on documentation 
			by NIST toolbox.

			Parameters
			----------
			T 	: float
				Atmospheric temperature in Kelvin

			Returns
			-------
			svp	: float
				Saturation vapour pressure
		'''
		K1 =  1.16705214528e+03
		K2 = -7.24213167032e+05
		K3 = -1.70738469401e+01
		K4 =  1.20208247025e+04
		K5 = -3.23255503223e+06
		K6 =  1.49151086135e+01
		K7 = -4.82326573616e+03
		K8 =  4.05113405421e+05
		K9 = -2.38555575678e-01
		K10 = 6.50175348448e+02
		_W = T + K9 / (T - K10)
		_A = _W**2 + K1*_W + K2
		_B = K3*_W**2 + K4*_W + K5
		_C = K6*_W**2 + K7*_W + K8
		_X = -_B + np.sqrt(_B**2 - 4*_A*_C)
		return (2*_C / _X)**4 * 1e6


	def xw(self, p, T, RH):
		'''
			Molar fraction of water vapour in moist air.

			Parameters
			----------
			p 	: float
				Atmospheric pressure in Pascal
			T 	: float
				Atmospheric temperature in Kelvin
			RH 	: float
				Atmospheric relative humidity (0<=RH<=1)

			Returns
			-------
			xw	: float
				Molar fraction of water vapour in moist air
		'''
		_f 		= self.f(p, T)
		_svp 	= self.svp(T)
		return _f * RH * _svp / p


	def Z(self, p, T, RH, xw=''):
		'''
			Compressibility of moist air

			Parameters
			----------
			p 	: float
				Atmospheric pressure in Pascal
			T 	: float
				Atmospheric temperature in Kelvin
			RH 	: float
				Atmospheric relative humidity (0<=RH<=1)

			Returns
			-------
			Z 	: float
				Compressibility of moist air
		'''
		_a0 =  1.58123e-6 # K Pa^-1
		_a1 = -2.9331e-8  # Pa^-1
		_a2 =  1.1043e-10 # K^-1 Pa^-1
		_b0 =  5.707e-6   # K Pa^-1
		_b1 = -2.051e-8   # Pa^-1
		_c0 =  1.9898e-4  # K Pa^-1
		_c1 = -2.376e-6   # Pa^-1
		_d  =  1.83e-11   # K^2 Pa^-2
		_e  = -0.765e-8   # K^2 Pa^-2
		_t  = T - 273.15
		if xw == '':
			_xw = self.xw(p, T, RH)
		else:
			_xw = xw
		Z = 1 - (p/T) * (_a0 + _a1*_t + _a2*_t**2 + (_b0 + _b1*_t)*_xw + (_c0 + _c1*_t)*_xw**2) + (p/T)**2 * (_d + _e*_xw**2)
		return Z


	def rho(self, p, T, RH, xc, xw=''):
		'''
			Calculates the density of air for the given parameters.
			
			Parameters
			----------
			p 	: float
				Atmospheric pressure in Pascal
			T 	: float
				Atmospheric temperature in Kelvin
			RH 	: float
				Atmospheric relative humidity (0<=RH<=1)
			xc 	: float
				CO2 density in parts per million
			xw 	: str or float
				Placeholder for molar fraction of water vapour in moist air.
				Used to pass the function along where it should go.

			Returns
			-------
			rho : float
				Density of air
		'''
		_Ma = 1e-3 * (28.9635 + 12.011e-6 * (xc - 400) )	# Molar mass in kg/mol of dry air containing xc ppm of CO2
		_Mw = 0.018015 										# kg/mol
		_R  = 8.314510										# J mol^-1 K^-1, gas constant
		if xw == '':
			_xw = self.xw(p, T, RH)
		else: 
			_xw = xw
		_Z  = self.Z(p, T, RH, xw=_xw)
		return ( ( p * _Ma ) / (_Z * _R * T) ) * ( 1 - _xw * (1 - _Mw/_Ma)) 


	def n_tph(self, l, T, p, RH, xc):
		'''
			Calculates the refraction of the atmosphere for any conditions.
			Follows the equations by Ciddor (1996).

			Parameters
			----------
			l 	: float
				Wavelength of light in microns
			p 	: float
				Atmospheric pressure in Pascal
			T 	: float
				Atmospheric temperature in Kelvin
			RH 	: float
				Atmospheric relative humidity (0<=RH<=1)
			xc 	: float
				CO2 density in parts per million

			Returns
			-------
			n 	: float
				Refractive index of air
		'''
		_n_a 	 = self.n_axs(l, xc)
		_n_w 	 = self.n_ws(l)
		_La 	 = (_n_a**2 - 1) / (_n_a**2 + 2)
		_Lw 	 = (_n_w**2 - 1) / (_n_w**2 + 2)
		
		# Calculate the density of the dry air components, using the partial pressure of the dry air, i.e. p_dry = p_atm - p_water
		_rho_a 	 = self.rho(p=p-RH*self.svp(T), T=T, RH=0, xc=xc)				# Dry air component for the actual conditions.
		_rho_axs = self.rho(p=101325, T=288.15, RH=0, xc=xc) 					# Dry air component at standard conditions.
		
		if RH == 0:
			# nan is returned for _rho_w if we put in a (partial) pressure of 0
			_rho_w = 0
		else:
			# Partial pressure of water vapour is p_water = h*svp
			_rho_w 	 = self.rho(p=RH*self.svp(T), T=T, RH=RH, xc=xc)			# Moist air component for the actual conditions.
		_rho_ws	 = self.rho(p=1333, T=293.15, RH=1333/self.svp(293.15), xc=xc) 	# Moist air component at standard conditions.
		
		_Ltotal  = ( _rho_a / _rho_axs ) * _La + ( _rho_w / _rho_ws ) * _Lw

		# print 'Dry air component = ', (_rho_a/_rho_axs)*(_n_a-1)*1e8
		# print 'Wet component = ', ( _rho_w / _rho_ws ) * (_n_w - 1)*1e8
		# print 'Sum of components = ', (_rho_a/_rho_axs)*(_n_a-1)*1e8+( _rho_w / _rho_ws ) * (_n_w - 1)*1e8
		return np.sqrt( (1 + 2 * _Ltotal) / (1 - _Ltotal) )


	def dn_tph(self, l, T, p, RH, xc, dl=None, dT=None, dP=None, dRH=None, dCO2=None):
		'''
			Calculates the uncertainty in the refractive index, following the Ciddor model,
			by using fully propagated error analysis.
			This involves almost endless use of the chain rule and evaluation of both the moist
			air and dry air components. Note: dry air does not contain any water, i.e. H=0, x_w=0, etc.

			Parameters
			----------
			l 	: float
				Wavelength of light in microns
			p 	: float
				Atmospheric pressure in Pascal
			T 	: float
				Atmospheric temperature in Kelvin
			RH 	: float
				Atmospheric relative humidity (0<=RH<=1)
			xc 	: float
				CO2 density in parts per million
			dl 	: float 
				Uncertainty in the wavelength in microns
			dT 	: float (default: 0.2)
				Uncertainty in the atmospheric temperature in Kelvin
			dP 	: float (default: 20)
				Uncertainty in the atmospheric pressure in Pascal
			dRH : float (default: 0.02)
				Uncertainty in the atmospheric relative humidity
			dCO2: float (default: 20)
				Uncertainty in the atmospheric CO2 density in parts per million

			Returns
			-------
			dn 	: float
				Uncertainty in the refractive index of air
		'''

		if dP == None:
			dP = self.dP
		if dT == None:
			dT = self.dT
		if dRH == None:
			dRH = self.dH
		if dCO2 == None:
			dCO2 = self.dCO2

		# constants
		_cf = 1.022
		_w0 = 295.235 	# um^-2
		_w1 = 2.6422	# um^-2
		_w2 = -0.032380	# um^-4
		_w3 = 0.004028	# um^-6

		_k0 = 238.0185 		# um^-2
		_k1 = 5792105  		# um^-2
		_k2 = 57.362		# um^-2
		_k3 = 167917		# um^-2	

		_A =  1.2378847e-5 # K^-2
		_B = -1.9121316e-2 # K^-2
		_C = 33.93711047
		_D = -6.3431645e3  # K^-2

		_Mw  = 0.018015   # kg/mol
		_R   = 8.314510	  # J mol^-1 K^-1, gas constant

		_a0 =  1.58123e-6 # K Pa^-1
		_a1 = -2.9331e-8  # Pa^-1
		_a2 =  1.1043e-10 # K^-1 Pa^-1
		_b0 =  5.707e-6   # K Pa^-1
		_b1 = -2.051e-8   # Pa^-1
		_c0 =  1.9898e-4  # K Pa^-1
		_c1 = -2.376e-6   # Pa^-1
		_d  =  1.83e-11   # K^2 Pa^-2
		_e  = -0.765e-8   # K^2 Pa^-2
		_t  = T - 273.15

		_aa = 1.00062
		_bb = 3.14e-8 	# Pa^-1
		_cc = 5.6e-7	# deg C^-2

		_RH_a = 0
		_RH_w = RH

		# Calculated values
		# _rho_obs = self.rho(p, T, RH, xc)
		# _ref	 = refraction(lat, h, p, _rho_obs)
		# _alpha   = _ref.rc * np.sin(np.deg2rad(z)) / (_ref.rc + _ref.H)
		_p_a 	 = p - RH*self.svp(T) 
		_p_w     = RH*self.svp(T)
		_x_w_a	 = self.xw(_p_a, T, _RH_a)
		_x_w_w 	 = self.f(_p_w, T)
		_Z_a 	 = self.Z(p=_p_a, T=T, RH=_RH_a)
		_Z_w 	 = self.Z(p=_p_w, T=T, RH=_RH_w, xw=_x_w_w)
		_Ma  	 = 1e-3 * (28.9635 + 12.011e-6 * (xc - 400) )	# Molar mass in kg/mol of dry air containing xc ppm of CO2
		_p_sv 	 = self.svp(T)
		_rho_a 	 = self.rho(p=p-RH*self.svp(T), T=T, RH=0, xc=xc)
		_rho_axs = self.rho(p=101325, T=288.15, RH=0, xc=xc)
		if RH == 0.0:
			# nan is returned for _rho_w if we put in a (partial) pressure of 0
			_rho_w = 0
		else:
			_rho_w = self.rho(p=RH*self.svp(T), T=T, RH=RH, xc=xc) 
		_rho_ws  = self.rho(p=1333, T=293.15, RH=1333/self.svp(293.15), xc=xc)
		_n_a1 = self.n_axs(l, xc)
		_n_w1 = self.n_ws(l)

		_L1_a = (_n_a1**2 - 1) / (_n_a1**2 + 2)  # Dry air component of L
		_L1_w = (_n_w1**2 - 1) / (_n_w1**2 + 2)
		_L1   = ( _rho_a / _rho_axs ) * _L1_a + ( _rho_w / _rho_ws ) * _L1_w # Overall L
		_n1   = np.sqrt( (1 + 2 * _L1) / (1 - _L1) )

		# Partial derivatives for derivation to l1
		_dn_wdl  = _cf * 1e-8 * (-1 * _w1 * 2/l**3 - _w2 * 4/l**5 - _w3 * 6/l**7)
		_dLdn_w1 = _rho_w / _rho_ws * 6 * _n_w1 / (_n_w1**2 + 2)**2
		_dn_adl  = 1e-8 * (1 + 0.534e-6 * (xc - 450)) * ( - (2 * _k1 * l ) / ((_k0 * l**2 - 1)**2) - (2 * _k3 * l) / ((_k2 * l**2 - 1)**2) )
		_dLdn_a1 = _rho_a / _rho_axs * 6 * _n_a1 / (_n_a1**2 + 2)**2
		_dndL1   = ( ( (1 + 2*_L1) / (1 - _L1)**2 ) + (2 / (1-_L1)) ) / (2 * np.sqrt( (1 + 2*_L1) / (1 - _L1) ) ) 
		_dndl   = _dndL1 * ( _dLdn_a1 * _dn_adl + _dLdn_w1 * _dn_wdl)

		# # Partial derivatives for derivation to l2


		# Partial derivatives for derivation to T
		# Dry air parts:
		_drho_adp_a  = _Ma / (_Z_a * _R * T) * (1 - _x_w_a * (1 - _Mw / _Ma))
		_dp_adp_sv 	 = - RH
		_dp_svdT 	 = (2*_A*T + _B - _D/(T**2)) * np.exp(_A * T**2 + _B * T + _C + _D/T )
		_drho_adTZ	 = - _p_a * _Ma / (_R * ( _Z_a * T)**2) * (1 - _x_w_a * (1 - _Mw / _Ma))
		_dTZdT_a	 = _p_a / T * (_a2 * _t**2 + _a1 * _t + _a0 + (_b1*_t + _b0) * _x_w_a + (_c0 + _c1*_t) * _x_w_a**2 ) - _p_a * (_a1 + 2*_a2*_t + _b1*_x_w_a + _c1*_x_w_a**2) - 2*(_p_a/T)**2 * (_d + _e * _x_w_a**2) + _Z_a
		_dTZdxw_a	 = 2*_e*_p_a**2 / T * _x_w_a - _p_a * (_b1*_t + _b0 + 2 * (_c0 + _c1*_t) * _x_w_a)
		_dxw_adf	 = _RH_a * _p_sv / _p_a
		_dfdT 		 = 2*_cc*_t
		_dxw_adp_sv	 = self.f(_p_a, T) * _RH_a / _p_a
		_drho_adxw_a = 0 # as rho_a has not water content, i.e. rho = p*M/ZRT


		# Moist air parts:
		_drho_wdp_w  = _Ma / (_Z_w * _R * T) * (1 - _x_w_w * (1 - _Mw / _Ma))
		_dp_wdp_sv 	 = RH
		_dp_svdT 	 = (2*_A*T + _B - _D/(T**2)) * np.exp(_A * T**2 + _B * T + _C + _D/T )
		_drho_wdTZ	 = - _p_w * _Ma / (_R * (_Z_w * T)**2) * (1 - _x_w_w * (1 - _Mw / _Ma))
		_dTZdT_w	 = _p_w / T * (_a2 * _t**2 + _a1 * _t + _a0 + (_b1*_t + _b0) * _x_w_w + (_c0 + _c1*_t) * _x_w_w**2 ) - _p_w * (_a1 + 2*_a2*_t + _b1*_x_w_w + _c1*_x_w_w**2) - 2*(_p_w/T)**2 * (_d + _e * _x_w_w**2) + _Z_w
		_dTZdxw_w	 = 2*_e*_p_w**2 / T * _x_w_w - _p_w * (_b1*_t + _b0 + 2 * (_c0 + _c1*_t) * _x_w_w)
		_dxw_wdf	 = 1
		_dfdT 		 = 2*_cc*_t
		_dxw_wdp_sv	 = self.f(_p_w, T) / _p_sv
		_drho_wdxw_w = _p_w * _Ma / (_Z_w * _R * T) * (_Mw / _Ma - 1)

		_tA =  _drho_adp_a * _dp_adp_sv * _dp_svdT +  _drho_adTZ * (_dTZdT_a + _dTZdxw_a * (_dxw_adf * _dfdT + _dxw_adp_sv * _dp_svdT) ) + _drho_adxw_a * ( _dxw_adf * _dfdT + _dxw_adp_sv * _dp_svdT )
		_tB =  _drho_wdp_w * _dp_wdp_sv * _dp_svdT +  _drho_wdTZ * (_dTZdT_w + _dTZdxw_w * (_dxw_wdf * _dfdT + _dxw_wdp_sv * _dp_svdT) ) + _drho_wdxw_w * ( _dxw_wdf * _dfdT + _dxw_wdp_sv * _dp_svdT )
		_dL1drho_a 	 = _L1_a / _rho_axs
		_dL1drho_w 	 = _L1_w / _rho_ws
		
		_dndT = _tA * (_dndL1 * _dL1drho_a) + _tB * (_dndL1 * _dL1drho_w)

		# Partial derivatives for derivation to p
		# Dry air parts:
		_drho_adZ 	= - _p_a * _Ma / ( _Z_a**2 * _R * T) * (1 - _x_w_a * (1 - _Mw / _Ma))
		_dZdp_a 	= - 1 / T * (_a0 + _a1 * _t + _a2 * _t**2 + (_b0 + _b1*_t)*_x_w_a + (_c0 + _c1*_t)*_x_w_a**2 ) + 2*_p_a/(T**2) * (_d + _e * _x_w_a**2)
		_dfdp 		= _bb
		_dxw_adp_a 	= - self.f(_p_a, T) * _RH_a * _p_sv / (_p_a)**2 

		# Moist air parts
		_dZdp_w 	= - 1 / T * (_a0 + _a1 * _t + _a2 * _t**2 + (_b0 + _b1*_t)*_x_w_w + (_c0 + _c1*_t)*_x_w_w**2 ) + 2*_p_w/(T**2) * (_d + _e * _x_w_w**2)
		# Not necessary, since dpwdp=0

		_pA = _drho_adp_a + _drho_adZ * _dZdp_a + _drho_adxw_a * (_dxw_adf * _dfdp + _dxw_adp_a)
		_dndP = _pA * (_dndL1 * _dL1drho_a)

		
		# Partial derivatives for derivation to H
		# Dry air parts:
		_dp_adH 	= - _p_sv
		_dZdxw_a 	= (_p_a / T)**2 * 2 * _e * _x_w_a - (_p_a / T) * ( (_b0 + _b1 * _t) + 2 * (_c0 + _c1 * _t) * _x_w_a )
		_dxw_adH 	= self.f(_p_a, T) * _p_sv / _p_a

		# Moist air parts:
		_dp_wdH 	= _p_sv
		_dZdxw_w 	= (_p_w / T)**2 * 2 * _e * _x_w_w - (_p_w / T) * ( (_b0 + _b1 * _t) + 2 * (_c0 + _c1 * _t) * _x_w_w )
		_dxw_wdH 	= 0
		_dxw_wdp_w 	= 0		
		_drho_wdZ 	= - _p_w * _Ma / ( _Z_w**2 * _R * T) * (1 - _x_w_w * (1 - _Mw / _Ma))

		_hA = _drho_adp_a * _dp_adH + _drho_adZ * ( _dZdp_a * _dp_adH + _dZdxw_a * (_dxw_adH + _dxw_adp_a * _dp_adH + _dxw_adf * _dfdp * _dp_adH) ) + _drho_adxw_a * (_dxw_adH + _dxw_adp_a * _dp_adH + _dxw_adf * _dfdp * _dp_adH)
		_hB = _drho_wdp_w * _dp_wdH + _drho_wdZ * ( _dZdp_w * _dp_wdH + _dZdxw_w * ( _bb * _p_sv ) ) + _drho_wdxw_w * ( _bb * _p_sv )
		_dndRH = _hA * (_dndL1 * _dL1drho_a) + _hB * (_dndL1 * _dL1drho_w)


		# Partial derivatives for derivation to xc
		# Dry air parts:
		_drho_adMa   = _p_a * (1 - _x_w_a) / (_Z_a * _R * T)
		_dL1drho_axs = - (_rho_a * _L1_a)/(_rho_axs**2)
		_drho_axsdMa = _drho_adMa
		_dL1dL_a	 = _rho_a / _rho_axs
		_dL_adn_a1	 = 6 * _n_a1 / (_n_a1**2 + 2)**2

		# Moist air parts:
		_drho_wdMa   = _p_w * (1 - _x_w_w) / (_Z_w * _R * T)
		_dL1drho_ws = - (_rho_w * _L1_w)/(_rho_ws**2)
		_drho_wsdMa = _drho_wdMa

		# Other parts:
		_dMadx_c   = 12.011e-9
		_dn_adx_c1 = (self.n_as(l) - 1) * 0.534e-6

		_dL1dx_c = _dL1drho_a * _drho_adMa * _dMadx_c + _dL1drho_w * _drho_wdMa * _dMadx_c + _dL1drho_axs * _drho_axsdMa * _dMadx_c + _dL1drho_ws * _drho_wsdMa * _dMadx_c + _dL1dL_a * _dL_adn_a1 * _dn_adx_c1
		_dndCO2 = _dndL1 * _dL1dx_c

		# Final error:
		_nError = np.sqrt( (_dndl * dl)**2 + (_dndT * dT)**2  + (_dndP * dP)**2 + (_dndRH * dRH)**2 + (_dndCO2 * dCO2)**2 )

		return _nError	


#!/usr/bin/env Python3
import numpy as np
from scipy import integrate
import os
from .ciddorModel import Observatory

'''
@ Author:		Joost van den Born 
@ Contact:		born@astron.nl
@ Description:	This script offers a complete model of the atmospheric refraction
				and dispersion. It uses the Ciddor1996 paper to calculate the 
				index of refraction for a wide range of conditions, with corresponding error propagation if needed.
				The actual refraction is calculated by any of the following models:
				- Plane parallel atmosphere
				- Cassini spherical homogeneous atmosphere
				- Mathar's Barometric exponential model
				- Other methods described in Corbard et al., 2019.
'''

class refraction(Observatory):
	'''
		Class that sets the observer location.
		These location attributes are then used to calculate the 
		appropriate atmospheric refraction.

		See also:
		Observatory()
		dispersion()
		...

		Attributes
		-----------
		lat : 	float
			Latitude of the observer in degrees
		h 	: 	float
			Altitude above mean sea level of the observer in meters
		Other Observatory() attributes

		Methods
		-------
		cassini(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric refraction assuming Cassini's homegeneous
			atmosphere model.
		cassiniError(z, l1, l2, T, p, RH, xc, dl1=0, dl2=0, dT=0.2, dP=20, 
					dRH=0.02, dCO2=20, dz=0, lat=None, h=None)
			Returns the uncertainty in the atmospheric refraction, assuming 
			Cassini's homegeneous atmosphere model and the given parameters.
		corbard(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric refraction following the method described 
			in Corbard et al., 2019.
		gammaCoefficients(n1, reduced_height=None)
			Helper function for matharExponential(), that calculates the gamma
			coefficients.
		H_isotherm(T)
			Sets the reduced height for the conditions at the observer, assuming
			an isothermal atmosphere.
		matharExponential(n1, zenith, reduced_height=None)
			Returns the atmospheric refraction assuming	Mathar's barometric 
			exponential model.
		oriani(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric refraction following the classic version
			of Oriani's theorem.	
		planeParallel(n1, n2, zenith)
			Returns the atmospheric refraction assuming a plane parallel
			atmosphere.
		psi(x)
			Submethod for corbard().
		refractionIntegral(n1, n2, zenith, R0=None, useStandardAtmosphere=True, heightData=None, rhoData=None)
			Returns the atmospheric refraction following the refraction integral.
		setReducedHeight(p, rho)
			Sets the reduced atmospheric height.
		set_H(p, rho)
			Sets the reduced height for the conditions at the observer, assuming
			an adiabatic atmosphere.
		set_rc()
			Sets the local radius of curvature of the Earth at the observer.			
		tan5(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric refraction following Oriani's theorem,
			including a fifth order term in the expansion.				
		Other Observatory() methods
	'''
	def __init__(self, lat, h):
		Observatory.__init__(self)
		self.h  	= h 					# height above sea level in meters
		self.lat 	= lat					# Latitude of the observer in degrees
		self.rc 	= self.set_rc() * 1000	# Radius of curvature of the earth at the observer
		# self.H2 	= self.H_isotherm(T) 	# Reduced height of isotherm atmosphere


	def cassini(self, n1, zenith, reduced_height=None):
		'''
			Refraction of spherical atmosphere, derived from geometry using Sine law

			Parameters
			----------
			n1 : float
				Refractive index at the wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			R : float
				Atmospheric refraction in degrees
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H
		
		_R1 = np.arcsin(n1 * self.rc * np.sin(np.deg2rad(zenith)) / (self.rc + H)) - np.arcsin(self.rc * np.sin(np.deg2rad(zenith)) / (self.rc + H) )
		return np.degrees(_R1)


	def cassiniError(self, n, dn, zenith, dz=0, reduced_height=None):
		'''
		Calculates the uncertainty in the refraction angle in degrees.
		See Observatory().n_tph() and Observatory().dn_tph() on how to
		calculate the refractive index and it's uncertainty.

		Parameters
		----------
		n 	: 	float
			Refractive index of air
		dn 	: 	float
			Uncertainty in the refractive index of air
		zenith : float
			Angle of observation in degrees.
			zenith = 0 means the observation is directly overhead.
		dz  : 	float (optional)
			Uncertainty in the zenith angle, in degrees.
		reduced_height : float (optional)
			Reduced height of the atmosphere. Default value is calculated
			from the object attributes.

		Returns
		-------
		dR 	: float
			Uncertainty in the atmospheric refraction in degrees.
		'''
		if reduced_height:
			_H = reduced_height
		else:
			_H = self.H
		_zenith = np.deg2rad(zenith)
		_dz 	= np.deg2rad(dz)
		_R 		= self.rc
		_dRdn 	= (_R * np.sin(_zenith)) / ((_R + _H) * np.sqrt(1 - ((n**2 * _R**2 * np.sin(_zenith)**2)/(_R + _H)**2)))

		_dRdz 	= (_R * np.cos(_zenith)) / (_R + _H) * ( n / np.sqrt(1-(n**2 * _R**2 * np.sin(_zenith)**2)/(_R + _H)**2) - 1 / np.sqrt(1-(_R**2 * np.sin(_zenith)**2)/(_R + _H)**2) )

		_dR 	= np.sqrt( (_dRdn * dn)**2 + (_dRdz * _dz)**2)

		return np.degrees(_dR)


	def corbard(self, n1, zenith, reduced_height=None):
		'''
			Corbard et al., 2019, mention an additional formula based on the error function.
			Oriani's theorem can be derived from this equation by 'keeping only the three first
			terms of its asymptotic expansion'.

			Parameters
			----------
			n1 : float
				Refractive index at the wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			R : float
				Atmospheric refraction in degrees
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H

		_a = n1 - 1
		_b = H / self.rc
		_R = _a * ( (2 - _a) / (np.sqrt( 2*_b - _a )) ) * np.sin( np.deg2rad(zenith) ) * self.psi( (np.cos(np.deg2rad(zenith))) / np.sqrt(2*_b - _a) )
		return np.degrees(_R)


	def gammaCoefficients(self, n1, reduced_height=None):
		'''
			Submethod for matharExponential(). Calculates the gamma
			coefficients.

			Parameters
			----------
			n1 	: 	float
				Refractive index at the wavelength of interest at the observer
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			g 	: 	array
				List of values, corresponding to the gamma coefficients.
				g[0] corresponds to g_1, g[1] to g_3, g[2] to g_5, and so on.
		'''

		if reduced_height:
			H = reduced_height
		else:
			H = self.H

		_chi0   = n1**2 - 1
		_Khat 	= H / self.rc

		_g1 = (1 - 5/24 * _Khat * _chi0**2 - 3/4 * _chi0 - 3/8 * _Khat**2 * _chi0 \
			- _Khat - 6 * _Khat**3 + 5/8 * _chi0**2 + 63/128 * _chi0**4 + 3/8 \
			* _Khat * _chi0 + 35/256 * _Khat * _chi0**3 + 2 * _Khat**2 + 5/36 \
			* _Khat**2 * _chi0**2 - 35/64 * _chi0**3 + 9/16 * _Khat**3 * _chi0 \
			- 120 * _Khat**5 + 24 * _Khat**4 - 231 / 512 * _chi0**5 - 63 / 640 \
			* _Khat * _chi0**4 - 35 / 512 * _Khat**2 * _chi0**3 - 5 / 36 * _Khat**3\
			* _chi0**2 - 9 / 8 * _Khat**4 * _chi0 )
		
		_g3 = (35/96 * _chi0**3 - 5/12 * _chi0**2 + 333 * _Khat**4 * _chi0 - 2400 \
			* _Khat**5 + 3605/4608 * _Khat**2 * _chi0**3 + 63/256 * _Khat * _chi0**4\
			 - 819/16 * _Khat**3 * _chi0 - 35/18 * _Khat**2 * _chi0**2 - 35/96 \
			 * _Khat * _chi0**3 + 69/8 * _Khat**2 * _chi0 + 5/8 * _Khat * _chi0**2 \
			 - 3/2 * _Khat * _chi0 + 1355/216 * _Khat**3 * _chi0**2 + 10 * _Khat**2 \
			 + 336 * _Khat**4 - 2 * _Khat + _chi0/2 - 21/64 * _chi0**4 - 54 * _Khat**3 \
			 + 77/256 * _chi0**5 ) * 1/2
		
		_g5 = (-96 * _Khat**3 + 8 * _Khat**2 - 7/24 * _chi0**3 + 65/4 * _Khat**2 \
			* _chi0**2 + 35/48 * _Khat * _chi0**3 + 30 * _Khat**2 * _chi0 - 5/3 \
			* _Khat * _chi0**2 - 3 * _Khat * _chi0 + 20091/8 * _Khat**4 * _chi0 \
			- 32605/216 * _Khat**3 * _chi0**2 - 35/9 * _Khat**2 * _chi0**3 + 984 \
			* _Khat**4 + _chi0**2/3 - 10200 * _Khat**5 - 77/320 * _chi0**5 + 21/80 \
			* _chi0**4 - 1089/4 * _Khat**3 * _chi0 - 7/16 * _Khat * _chi0**4 ) * 3/8
		
		_g7 = (-17520 * _Khat**5 - 9/40 * _chi0**4 - 8755/12 * _Khat**3 * _chi0**2 \
			+ 1056 * _Khat**4 + _chi0**3/4 + 23517/4 * _Khat**4 * _chi0 - 11/3 \
			* _Khat * _chi0**2 + 21 * _Khat**2 * _chi0 - 7/4 * _Khat * _chi0**3 \
			+ 505/9 * _Khat**2 * _chi0**2 - 801/2 * _Khat**3 * _chi0 + 33/160 \
			* _chi0**5 - 48 * _Khat**3 + 63/80 * _Khat * _chi0**4 + 2345/96 \
			* _Khat**2 * _chi0**3) * 5/16
		
		_g9 = (_chi0**4 / 5 + 384 * _Khat**4 - 13440 * _Khat**5 + 5562 * _Khat**4 \
			* _chi0 + 3125/36 * _Khat**2 * _chi0**3 - 11/60 * _chi0**5 - 25/6 \
			* _Khat * _chi0**3 - 26950/27 * _Khat**3 * _chi0**2 + 340/9 * _Khat**2 \
			* _chi0**2 - 180 * _Khat**3 * _chi0 - 9/5 * _Khat * _chi0**4) * 35/128

		g = np.array([_g1, _g3, _g5, _g7, _g9])
		g*= n1*_chi0/2

		return g		
	

	def H_isotherm(self, T):
		'''
			Calculates the reduced height of the atmosphere, 
			assuming an isotherm distribution.

			Parameters
			----------
			T 	: float
				Atmospheric temperature in Kelvin

			Returns
			-------
			H 	: float
				Reduced height of the atmosphere in meters
		'''
		_kb = 1.38064852e-23 # kg m2 s-2 K-1
		_m  = 4.809651698e-26 # kg (weight per molecule of air, assuming dry air)

		_phi = np.deg2rad(self.lat)
		_c1 = 5.2790414e-3
		_c2 = 2.32718e-5
		_c3 = 1.262e-7
		_c4 = 7e-10
		_g0 = 9.780327 # m/s^2
		_g0_local = _g0 * (1 + _c1 * np.sin(_phi)**2 + _c2 * np.sin(_phi)**4 + _c3 * np.sin(_phi)**6 + _c4 * np.sin(_phi)**8)

		return (_kb*T) / (_m * _g0_local)


	def matharExponential(self, n1, zenith, reduced_height=None):
		'''
			Model on atmospheric refraction described by Richard Mathar in 
			"A Barometric Exponential Model of the Atmosphere’s Refractive Index:  ZenithAngles and Second Order Aberration in the Entrance Pupil" (2020)
			ArXiv id: 2004.11808 (https://arxiv.org/abs/2004.11808)
			
			The code in this function is adapted from eqs 3, 16, 17 and 18 in this
			work. It is an adaptation of the RefrExpo.cxx C++ file written by Mathar.

			This model assumes the refractive index, n, is related to the susceptibility, 
			chi, following: n(r) = sqrt(1+chi(r)) (eq. 9).

			R. Mathar then assumes that the susceptibility is proportional to the
			air density and reduced height, following eq. 10:
			chi(r) = chi0 * exp(-(r-rho)/K).

			The refraction is described using a Taylor expansion; R in terms of
			the odd powers of the tangent of the observed zenith angle (eq. 3), not
			unlike Oriani's theorem. This requires coefficients g_i.

			Parameters
			----------
			n1 : float
				Refractive index at the wavelength of interest at the observer
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			R : float
				Atmospheric refraction in degrees
		'''

		if reduced_height:
			_H = reduced_height
		else:
			_H = self.H

		_zenith	= np.deg2rad(zenith)
		_rc 	= self.rc # rho

		# Atmospheric refraction in radians
		_gamma  = self.gammaCoefficients(n1, reduced_height=_H)
		_R = sum([_gamma[i]*np.tan(_zenith)**(2*i+1) for i in range(len(_gamma))])
		
		return np.degrees(_R)	


	def oriani(self, n1, zenith, reduced_height=None):
		'''
			Classic refraction formula with two tan(z) terms. This does not 
			assume any information about the structure of the atmosphere. 
			Found with Oriani's theorem.

			Parameters
			----------
			n1 : float
				Refractive index at the wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			R : float
				Atmospheric refraction in degrees
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H

		_a = n1 - 1
		_b = H / self.rc
		_R = _a * (1 - _b) * np.tan( np.deg2rad(zenith) ) - _a * (_b - _a/2) * np.tan( np.deg2rad(zenith) )**3
		return np.degrees(_R)


	def planeParallel(self, n1, zenith):
		'''
			Refraction based on a flat atmosphere, easily derived from Snell's law.

			Parameters
			----------
			n1 : float
				Refractive index at the wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.

			Returns
			-------
			R : float
				Atmospheric refraction in degrees
		'''
		_R = np.arcsin(n1 * np.sin( np.deg2rad(zenith) )) - np.deg2rad(zenith)
		return np.degrees(_R)


	def psi(self, x):
		'''
			Submethod for corbard(). Described explicitly in Corbard et al.,2019.

			Parameters
			----------
			x 	: 	float
				Input of the function

			Returns
			-------
			y 	: 	float
				Output of the function
		'''
		_f   = lambda a: np.exp(-a**2)
		_int = integrate.quad(_f, x, np.inf)
		return np.exp(x**2) * _int[0]


	def refractionIntegral(self, n1, zenith, R0=None, useStandardAtmosphere=True,
		heightData=None, rhoData=None):
		'''
			Using data that gives temperature, pressure and density as a 
			function of height, the gladstone-dale relation is invoked to
			calculate the path of a light ray along the atmosphere. 
			This is allows us to also include atmospheric activity in 
			different atmosphere layers.
			Recommended use with the US Standard Atmosphere from 1976.
			USSA1976 data generated from code supplied by http://www.pdas.com/atmos.html
			The method described in eq. 3 in Auer & Standish (2000) is used for the 
			calculation.

			Parameters
			----------
			n1 : float
				Refractive index at the wavelength of interest at the observer
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			R0 : float (optional)
				Local radius of curvature of the Earth in km.
			useStandardAtmosphere : bool
				Flag to use the included USSA1976 data
			heightData : 1D Array
				If "useStandardAtmosphere" is False, the function will ask
				for alternative atmospheric data to integrate over. This
				gives the altitudinal points
			rhoData : 1D Array
				If useStandardAtmosphere is False, the function will ask
				for alternative atmospheric data to integrate over. This
				gives the atmospheric density at the points defined by 
				"heightData".

			Returns
			-------
			R : float
				Atmospheric refraction in degrees

		'''
		if not R0:
			R0 = self.rc
		if useStandardAtmosphere:
			_datafileLocation = os.path.join(os.path.dirname(__file__), 'data/1976USSA.txt')
			_h, _, _, _, _Temp, _Press, _Dens,_,_,_ = np.genfromtxt(_datafileLocation, unpack=True, skip_header=6)
		else:
			_Dens = rhoData
			_h 	  = heightData

		zenith 		= np.deg2rad(zenith)
		_Dens 	= np.asarray(_Dens)
		_h 	 	= np.asarray(_h) 
		_R 		= R0 + _h * 1000
		
		# Gladstone-Dale relation between refractive index and density
		_n_h = 1 + (n1 - 1) / _Dens[0] * _Dens 

		# Calculation of the zenith angle for given height and refractive index, from the refractive invariant
		_z 		= np.arcsin( (n1 * R0) / (_n_h * _R) * np.sin(zenith) )
		
		# Calculation of log space derivative of n and r
		_lnn = np.log(_n_h)
		_lnr = np.log(_R)
		_dlnn = np.asarray([_lnn[i+1] - _lnn[i] for i in range(len(_lnn)-1)])
		_dlnr = np.asarray([_lnr[i+1] - _lnr[i] for i in range(len(_lnr)-1)])
		_dz   = np.asarray([(_z[i] - _z[i+1]) for i in range(len(_z)-1)])

		# Calculation of the integrand in eq. 3 in Auer & Standish
		_Integrand = -1*(_dlnn/_dlnr) / (1 + _dlnn/_dlnr) * _dz


		return np.degrees(np.sum(_Integrand))


	def setReducedHeight(self, p, rho):
		'''
			Sets the reduced atmospheric height.

			Parameters
			----------
			p 	: float
				Atmospheric pressure in Pascal
			rho : float
				Density of the atmospheric air at the observer
		'''
		self.H  	= self.set_H(p, rho) 		# Reduced height of the atmosphere assuming ideal gas law
	

	def set_H(self, p, rho):
		'''
			Calculates the reduced height of the atmosphere, assuming ideal gas law.

			Parameters
			----------
			p 	: float
				Atmospheric pressure in Pascal
			rho : float
				Density of the atmospheric air at the observer

			Returns
			-------
			H 	: float
				Reduced height of the atmosphere in meters
		'''
		_phi = np.deg2rad(self.lat)
		_c1 = 5.2790414e-3
		_c2 = 2.32718e-5
		_c3 = 1.262e-7
		_c4 = 7e-10
		_g0 = 9.780327 # m/s^2
		_g0_local = _g0 * (1 + _c1 * np.sin(_phi)**2 + _c2 * np.sin(_phi)**4 + _c3 * np.sin(_phi)**6 + _c4 * np.sin(_phi)**8)
		_g = _g0_local - (3.0877e-6 - 4.3e-9 * np.sin(_phi)**2) * self.h + 7.2e-13 * self.h**2
		return p / (rho * _g)


	def set_rc(self):
		'''
			Returns the radius of the curvature at the observer in km,
			taking into account the latitude of the observer and the corresponding
			ellipsoidality of the earth.

			Returns
			-------
			rc 	: float
				Local radius of curvature of the Earth in kilometers.
		'''
		_A   = 6378.137 # km
		_B   = 6356.752 # km
		_phi = np.deg2rad(self.lat)
		
		_rc0 = (_A*_B)**2 / (_A**2 * np.cos(_phi)**2 + _B**2 * np.sin(_phi)**2)**(3/2)
		return _rc0 + self.h/1000


	def tan5(self, n1, zenith, reduced_height=None):
		'''
			Same as oriani's theorem, but including a higher order term.
			Found in Corbard et al., 2019.

			Parameters
			----------
			n1 : float
				Refractive index at the wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			R : float
				Atmospheric refraction in degrees
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H
		
		_a = n1 - 1
		_b = H / self.rc
		_R = _a * (1 - _b) * np.tan( np.deg2rad(zenith) )  - _a * (_b - _a/2) * np.tan( np.deg2rad(zenith) )**3 + 3 * _a * (_b - _a/2)**2 * np.tan( np.deg2rad(zenith) )**5
		return np.degrees(_R)




#!/usr/bin/env Python3
import numpy as np

def sla_atms(RT, TT, DNT, GAMAL, R):
	'''
		Internal routine used by sla_refro(), ported from the SLALIB 
		FORTRAN code (Copyright P.T.Wallace.  All rights reserved.)

		The method is also described in "Explanatory Supplement 
		Astronomical Almanac" by P.K. Seidelmann, 1992, section 3.281.

		Refractive index and derivative with respect to height for the
		stratosphere.

		Parameters
		----------
		RT 		: 	float
			Height of tropopause from center of the Earth (meters)
		TT 		: 	float
			Temperature at the tropopause (K)
		DNT     : 	float
			Refractive index at the tropopause
		GAMAL 	: 	float
			Constant of the atmospheric model = G*MD/R
		R 		: 	float
			Current distance from the centre of the Earth (metre)

		Returns
		-------
		DN 		: 	float
			Refractive index at R
		RDNDR 	: 	float
			R * rate the refractive index is changing at R
	'''
	_B 		= GAMAL / TT
	_W 		= (DNT - 1) * np.exp( -_B * (R - RT) )
	DN  	= 1+_W
	RDNDR 	= -R * _B * _W
	return DN, RDNDR


def sla_atmt(R0, T0, ALPHA, GAMM2, DELM2, C1, C2, C3, C4, C5, C6, R):
	'''
		Internal routine used by sla_refro(), ported from the SLALIB 
		FORTRAN code (Copyright P.T.Wallace.  All rights reserved.)

		The method is also described in "Explanatory Supplement 
		Astronomical Almanac" by P.K. Seidelmann, 1992, section 3.281.

		Refractive index and derivative with respect to height for the
		troposphere.

		Parameters
		----------
		R0 		: 	float
			Height of observer from center of the Earth (meter)
		T0 		: 	float
			Temperature at the observer (K)
		ALPHA 	: 	float
			Alpha
		GAMM2 	: 	float 
			Gamma minus 2 - see HMNAO paper
		DELM2 	: 	float
			Delta minus 2
		C1  	: 	float
			Useful term
		C2  	: 	float
			Useful term
		C3  	: 	float
			Useful term
		C4  	: 	float
			Useful term
		C5  	: 	float
			Useful term
		C6  	: 	float
			Useful term
		R    	:  	float
			Current distance from the center of the Earth (meters)

		Returns
		-------
		T 		: 	float
			Temperature at R (K)
		DN 		:   float 
			Refractive index at R
		RDNDR 	: 	float
			R * rate the refractive index is changing at R

		Note that in the optical case C5 and C6 are zero.
	'''
	T 		= max(min(T0 - ALPHA * (R - R0), 320), 100)
	TT0 	= T / T0
	TT0GM2 	= TT0**GAMM2
	TT0DM2 	= TT0**DELM2
	DN 		= 1 + ( C1 * TT0GM2 - (C2 - C5 / T) * TT0DM2) * TT0
	RDNDR 	= R * (-C3 * TT0GM2 + (C4 - C6 / TT0) * TT0DM2)
	return T, DN, RDNDR


def sla_drange(a):
	'''
		Helper function for sla_refro().
	'''
	drange = a % (2*np.pi)
	if drange >= np.pi:
		drange -= np.abs(2*np.pi)*np.sign(a)
	return drange


def sla_refi(dn, rdndr):
	'''
		Helper function for sla_refro(), that calculates
		the refraction integrand.
	'''
	return rdndr/(dn+rdndr)

def sla_refro(ZOBS, HM, TDK, PMB, RH, WL, PHI, TLR, EPS):
	'''
		This is a direct port from the original FORTRAN SLA_REFRO()
		routine.

		Calculates Atmospheric refraction for radio and 
		optical/IR wavelengths, ported from the original SLALIB
		FORTRAN code (Copyright P.T.Wallace.  All rights reserved.).

		The method is also described in "Explanatory Supplement 
		Astronomical Almanac" by P.K. Seidelmann, 1992, section 3.281.
		
		Parameters
		----------
		ZOBS 	:	float
			Observed zenith distance of the source (radian)
		HM 		:	float
			Height of the observer above sea level (meter)
		TDK 	:	float
			Ambient temperature at the observer (K)
		PMB 	:	float
			Pressure at the observer (millibar)
		RH 		:	float
			Relative humidity at the observer (range 0-1)
		WL 		:	float
			Effective wavelength of the source (micrometre)
		PHI 	:	float
			Latitude of the observer (radian, astronomical)
		TLR 	:	float
			Temperature lapse rate in the troposphere (K/metre)
		EPS 	:	float
			Precision required to terminate iteration (radian)
		
		Returns
		-------
		REF 	:	float
			Refraction: in vacuo ZD minus observed ZD (radian)
		
		Notes
		-----
		1  A suggested value for the TLR argument is 0.0065D0.  The
		   refraction is significantly affected by TLR, and if studies
		   of the local atmosphere have been carried out a better TLR
		   value may be available.  The sign of the supplied TLR value
		   is ignored.
		2  A suggested value for the EPS argument is 1D-8.  The result is
		   usually at least two orders of magnitude more computationally
		   precise than the supplied EPS value.
		3  The routine computes the refraction for zenith distances up
		   to and a little beyond 90 deg using the method of Hohenkerk
		   and Sinclair (NAO Technical Notes 59 and 63, subsequently adopted
		   in the Explanatory Supplement, 1992 edition - see section 3.281).
		4  The code is a development of the optical/IR refraction subroutine
		   AREF of C.Hohenkerk (HMNAO, September 1984), with extensions to
		   support the radio case.  Apart from merely cosmetic changes, the
		   following modifications to the original HMNAO optical/IR refraction
		   code have been made:
		   .  The angle arguments have been changed to radians.
		   .  Any value of ZOBS is allowed (see note 6, below).
		   .  Other argument values have been limited to safe values.
		   .  Murray's values for the gas constants have been used
		      (Vectorial Astrometry, Adam Hilger, 1983).
		   .  The numerical integration phase has been rearranged for
		      extra clarity.
		   .  A better model for Ps(T) has been adopted (taken from
		      Gill, Atmosphere-Ocean Dynamics, Academic Press, 1982).
		   .  More accurate expressions for Pwo have been adopted
		      (again from Gill 1982).
		   .  The formula for the water vapour pressure, given the
		      saturation pressure and the relative humidity, is from
		      Crane (1976), expression 2.5.5.
		   .  Provision for radio wavelengths has been added using
		      expressions devised by A.T.Sinclair, RGO (private
		      communication 1989).  The refractivity model currently
		      used is from J.M.Rueger, "Refractive Index Formulae for
		      Electronic Distance Measurement with Radio and Millimetre
		      Waves", in Unisurv Report S-68 (2002), School of Surveying
		      and Spatial Information Systems, University of New South
		      Wales, Sydney, Australia.
		   .  The optical refractivity for dry air is from Resolution 3 of
		      the International Association of Geodesy adopted at the XXIIth
		      General Assembly in Birmingham, UK, 1999.
		   .  Various small changes have been made to gain speed.
		5  The radio refraction is chosen by specifying WL > 100 micrometres.
		   Because the algorithm takes no account of the ionosphere, the
		   accuracy deteriorates at low frequencies, below about 30 MHz.
		6  Before use, the value of ZOBS is expressed in the range +/- pi.
		   If this ranged ZOBS is -ve, the result REF is computed from its
		   absolute value before being made -ve to match.  In addition, if
		   it has an absolute value greater than 93 deg, a fixed REF value
		   equal to the result for ZOBS = 93 deg is returned, appropriately
		   signed.
		7  As in the original Hohenkerk and Sinclair algorithm, fixed values
		   of the water vapour polytrope exponent, the height of the
		   tropopause, and the height at which refraction is negligible are
		   used.
		8  The radio refraction has been tested against work done by
		   Iain Coulson, JACH, (private communication 1995) for the
		   James Clerk Maxwell Telescope, Mauna Kea.  For typical conditions,
		   agreement at the 0.1 arcsec level is achieved for moderate ZD,
		   worsening to perhaps 0.5-1.0 arcsec at ZD 80 deg.  At hot and
		   humid sea-level sites the accuracy will not be as good.
		9  It should be noted that the relative humidity RH is formally
		   defined in terms of "mixing ratio" rather than pressures or
		   densities as is often stated.  It is the mass of water per unit
		   mass of dry air divided by that for saturated air at the same
		   temperature and pressure (see Gill 1982).
		10 The algorithm is designed for observers in the troposphere.  The
		   supplied temperature, pressure and lapse rate are assumed to be
		   for a point in the troposphere and are used to define a model
		   atmosphere with the tropopause at 11km altitude and a constant
		   temperature above that.  However, in practice, the refraction
		   values returned for stratospheric observers, at altitudes up to
		   25km, are quite usable.
	'''

	# Fixed parameters
	D93 	= np.deg2rad(93)# 93 degrees in radians
	GCR 	= 8314.32 		# Universal gas constant
	DMD 	= 28.9644 		# Molecular weight of dry air
	DMW 	= 18.0152 		# Molecular weight of water vapour
	S 		= 6378120 		# Mean Earth radius (metre)
	DELTA 	= 18.36 		# Exponent of temperature dependence of water vapour pressure
	HT 		= 11000			# Height of tropopause (metre)
	HS 		= 80000			# Upper limit for refractive effects (metre)
	ISMAX 	= 16384			# Numerical integration: maximum number of strips.
	
	# Transform ZOBS into the normal range.
	ZOBS1 = sla_drange(ZOBS)
	ZOBS2 = min(abs(ZOBS1),D93)

	# Keep other arguments within safe bounds.
	HMOK  = min(max(HM,-1E3),HS)
	TDKOK = min(max(TDK,100),500)
	PMBOK = min(max(PMB,0),10000)
	RHOK  = min(max(RH,0),1)
	WLOK  = max(WL,0.1)
	ALPHA = min(max(abs(TLR),0.001),0.01)

	# Tolerance for iteration.
	TOL   = min(max(abs(EPS),1E-12),0.1)/2

	# Decide whether optical/IR or radio case - switch at 100 microns.
	OPTIC = (WLOK<=100)

	# Set up model atmosphere parameters defined at the observer.
	WLSQ  = WLOK*WLOK

	# Local gravity, can be made consistent with AstroAtmosphere
	GB    = 9.784*(1-0.0026*np.cos(PHI+PHI)-0.00000028*HMOK)

	if OPTIC:
		# Refractivity from Barrell & Sears (1939)!
		A = (287.6155+(1.62887+0.01360/WLSQ)/WLSQ)*273.15E-6/1013.25
	else:
		A = 77.6890E-6

	GAMAL = (GB*DMD)/GCR
	GAMMA = GAMAL/ALPHA
	GAMM2 = GAMMA-2
	DELM2 = DELTA-2
	TDC   = TDKOK-273.15
	PSAT  = 10**((0.7859+0.03477*TDC)/(1+0.00412*TDC))\
				*(1+PMBOK*(4.5E-6+6E-10*TDC*TDC))

	if PMBOK > 0:
		PWO = RHOK*PSAT/(1-(1-RHOK)*PSAT/PMBOK)
	else:
		PWO = 0

	W  = PWO*(1-DMW/DMD)*GAMMA/(DELTA-GAMMA)
	C1 = A*(PMBOK+W)/TDKOK
	if OPTIC:
		C2 = (A*W+11.2684E-6*PWO)/TDKOK
	else:
		C2 = (A*W+6.3938E-6*PWO)/TDKOK
	
	C3 = (GAMMA-1)*ALPHA*C1/TDKOK
	C4 = (DELTA-1)*ALPHA*C2/TDKOK
	if OPTIC:
		C5 = 0
		C6 = 0
	else:
		C5 = 375463E-6*PWO/TDKOK
		C6 = C5*DELM2*ALPHA/(TDKOK*TDKOK)

	# Conditions at the observer.
	R0  = S+HMOK
	TEMPO, DN0, RDNDR0 = sla_atmt(R0,TDKOK,ALPHA,GAMM2,DELM2,C1,C2,C3,C4,C5,C6,R0)
	SK0 = DN0*R0*np.sin(ZOBS2)
	F0  = sla_refi(DN0,RDNDR0)

	# Conditions in the troposphere at the tropopause.
	RT = S+max(HT,HMOK)
	TT, DNT, RDNDRT = sla_atmt(R0,TDKOK,ALPHA,GAMM2,DELM2,C1,C2,C3,C4,C5,C6,RT)
	SINE = SK0/(RT*DNT)
	ZT = np.arctan2(SINE,np.sqrt(max(1-SINE*SINE,0)))
	FT = sla_refi(DNT,RDNDRT)

	# Conditions in the stratosphere at the tropopause.
	DNTS,RDNDRP = sla_atms(RT,TT,DNT,GAMAL,RT)
	SINE = SK0/(RT*DNTS)
	ZTS  = np.arctan2(SINE,np.sqrt(max(1-SINE*SINE,0)))
	FTS  = sla_refi(DNTS,RDNDRP)

	# Conditions at the stratosphere limit.
	RS   = S+HS
	DNS,RDNDRS = sla_atms(RT,TT,DNT,GAMAL,RS)
	SINE = SK0/(RS*DNS)
	ZS   = np.arctan2(SINE,np.sqrt(max(1-SINE*SINE,0)))
	FS   = sla_refi(DNS,RDNDRS)

	# Variable initialization to avoid compiler warning.
	REFT = 0

	# Integrate the refraction integral in two parts;  first in the
	# troposphere (K=1), then in the stratosphere (K=2).

	for K in [1,2]:
		# Initialize previous refraction to ensure at least two iterations.
		REFOLD = 1

		# Start off with 8 strips.
		IS = 8

		# Start Z, Z range, and start and end values.
		if K==1:
			Z0 = ZOBS2
			ZRANGE = ZT-Z0
			FB = F0
			FF = FT
		else:
			Z0 = ZTS
			ZRANGE = ZS-Z0
			FB = FTS
			FF = FS

		# Sums of odd and even values.
		FO = 0
		FE = 0

		# First time through the loop we have to do every point.
		N = 1

		# Start of iteration loop (terminates at specified precision).
		LOOP = True
		while LOOP:

			# Strip width.
			H = ZRANGE/IS

			# Initialize distance from Earth centre for quadrature pass.
			if K==1:
				R = R0
			else:
				R = RT

			# One pass (no need to compute evens after first time).
			# DO I=1,IS-1,N
			for I in np.arange(1, IS-1, N):

				# Sine of observed zenith distance.
				SZ = np.sin(Z0+H*I)

				# Find R (to the nearest metre, maximum four iterations).
				if (SZ>1E-20):
					W  = SK0/SZ
					RG = R
					DR = 1E6
					J  = 0
					while (abs(DR)>1 and J<4):
						J+=1
						if K==1:
							TG,DN,RDNDR = sla_atmt(R0,TDKOK,ALPHA,GAMM2,DELM2,C1,C2,C3,C4,C5,C6,RG)
						else:
							DN,RDNDR = sla_atms(RT,TT,DNT,GAMAL,RG)
						DR = (RG*DN-W)/(DN+RDNDR)
						RG = RG-DR
					R = RG

				# Find the refractive index and integrand at R.
				if K==1:
					T,DN,RDNDR = sla_atmt(R0,TDKOK,ALPHA,GAMM2,DELM2,C1,C2,C3,C4,C5,C6,R)
				else:
					DN,RDNDR = sla_atms(RT,TT,DNT,GAMAL,R)

				F = sla_refi(DN,RDNDR)

				# Accumulate odd and (first time only) even values.
				if N==1 and (I%2)==0:
					FE = FE+F
				else:
					FO = FO+F


			# Evaluate the integrand using Simpson's Rule.
			REFP = H*(FB+4*FO+2*FE+FF)/3

			# Has the required precision been achieved (or can't be)?
			if (abs(REFP-REFOLD)>TOL and IS<ISMAX):
				# No: prepare for next iteration.

				# Save current value for convergence test.
				REFOLD = REFP

				# Double the number of strips.
				IS = IS+IS

				# Sum of all current values = sum of next pass's even values.
				FE = FE+FO

				# Prepare for new odd values.
				FO = 0

				# Skip even values next time.
				N = 2
			else:
				# Yes: save troposphere component and terminate the loop.
				if (K==1):
					REFT = REFP
				LOOP = False

	# Result.
	REF = REFT+REFP
	if (ZOBS1<0): 
		REF = -REF
	return REF


# =============================================================================
# WORK IN PROGRESS; Implementation of the slalib model for 
# 					any refractive index model.
# Main issue is in the calculation of dn/dr for any model
# of the refractive index. One possibility is following 
# the Corbard error function model, which assumes the
# atmospheric density follows the equation rho(r)=exp(-u/a).
# But then the SLA_REFRO routine would not have to be modified
# and one would simply use the refraction.corbard() function.
# =============================================================================
#
# def get_rc(lat):
# 		'''
# 			Returns the radius of the curvature at the observer in km,
# 			taking into account the latitude of the observer and the corresponding
# 			ellipsoidality of the earth.

# 			Returns
# 			-------
# 			rc 	: float
# 				Local radius of curvature of the Earth in kilometers.
# 		'''
# 		_A   = 6378.137 # km
# 		_B   = 6356.752 # km
# 		_phi = np.deg2rad(lat)
		
# 		_rc0 = (_A*_B)**2 / (_A**2 * np.cos(_phi)**2 + _B**2 * np.sin(_phi)**2)**(3/2)
# 		return _rc0*1000


# def get_local_g(lat, h):
# 	'''
# 		Returns the local gravitic acceleration for a given latitude 
# 		and altitude.

# 		Parameters
# 		----------
# 		lat : float
# 			Latitude in degrees
# 		h 	: float
# 			Altitude above sea level in meters

# 		Returns
# 		-------
# 		g 	: float
# 			Local acceleration due to gravity in m/s^2
# 	'''
# 	_phi = np.deg2rad(lat)
# 	_c1 = 5.2790414e-3
# 	_c2 = 2.32718e-5
# 	_c3 = 1.262e-7
# 	_c4 = 7e-10
# 	_g0 = 9.780327 # m/s^2
# 	_g0_local = _g0 * (1 + _c1 * np.sin(_phi)**2 + _c2 * np.sin(_phi)**4 + _c3 * np.sin(_phi)**6 + _c4 * np.sin(_phi)**8)
# 	_g = _g0_local - (3.0877e-6 - 4.3e-9 * np.sin(_phi)**2) * h + 7.2e-13 * h**2
# 	return _g

# def aa_refro(zenith, l, h, T, p, RH, xc, lat, TLR, EPS):
# 	'''
# 		Modified version of sla_refro(), updating some constants
# 		and changing the units on input and output parameters.

# 		The function calculates the atmospheric refraction
# 		optical/IR wavelengths (not radio!), ported from the original 
# 		SLALIB FORTRAN code (Copyright P.T.Wallace.  All rights 
# 		reserved.).

# 		The method is also described in "Explanatory Supplement 
# 		Astronomical Almanac" by P.K. Seidelmann, 1992, section 3.281.

# 		See also the description of sla_refro().
		
# 		Parameters
# 		----------
# 		zenith 	:	float
# 			Observed zenith distance of the source (degrees)
# 		l 		:	float
# 			Effective wavelength of the source (micrometre)
# 		h 		:	float
# 			Height of the observer above sea level (meter)
# 		T 		:	float
# 			Ambient temperature at the observer (K)
# 		p 		:	float
# 			Pressure at the observer (Pascal)
# 		RH 		:	float
# 			Relative humidity at the observer (range 0-1)
# 		xc 		: float
# 			CO2 density in parts per million
# 		lat 	:	float
# 			Latitude of the observer (degrees)
# 		TLR 	:	float
# 			Temperature lapse rate in the troposphere (K/metre)
# 		EPS 	:	float
# 			Precision required to terminate iteration (radian)
		
# 		Returns
# 		-------
# 		REF 	:	float
# 			Refraction: in vacuo ZD minus observed ZD (radian)
# 	'''
	
# 	# Converting some units for use later
# 	ZOBS = np.deg2rad(zenith)
# 	HM 	 = h
# 	TDK  = T
# 	PMB  = p / 100
# 	WL   = l
# 	PHI  = np.deg2rad(lat)

# 	# Fixed parameters
# 	D93 	= np.deg2rad(93)# 93 degrees in radians
# 	GCR 	= 8314.510 		# Universal gas constant
# 	DMD 	= 28.9635 + 12.011e-6 * (xc - 400) # Molar mass in g/mol of dry air containing xc ppm of CO2
# 	DMW 	= 18.0152 		# Molecular weight of water vapour
# 	S 		= get_rc(lat)	# Mean Earth radius (meters)
# 	DELTA 	= 18.36 		# Exponent of temperature dependence of water vapour pressure
# 	HT 		= 11000			# Height of tropopause (metre)
# 	HS 		= 80000			# Upper limit for refractive effects (meters)
# 	ISMAX 	= 16384			# Numerical integration: maximum number of strips.
	
# 	# Transform ZOBS into the normal range.
# 	ZOBS1 = sla_drange(ZOBS)
# 	ZOBS2 = min(abs(ZOBS1),D93)

# 	# Keep other arguments within safe bounds.
# 	HMOK  = min(max(HM,-1E3),HS)
# 	TDKOK = min(max(TDK,100),500)
# 	PMBOK = min(max(PMB,0),10000)
# 	RHOK  = min(max(RH,0),1)
# 	WLOK  = max(WL,0.1)
# 	ALPHA = min(max(abs(TLR),0.001),0.01)

# 	# Tolerance for iteration.
# 	TOL   = min(max(abs(EPS),1E-12),0.1)/2

# 	# Set up model atmosphere parameters defined at the observer.
# 	WLSQ  = WLOK*WLOK

# 	# Local gravity
# 	GB = get_local_g(lat, h)

# 	# Refractivity of air
# 	A = (287.6155+(1.62887+0.01360/WLSQ)/WLSQ)*273.15E-6/1013.25

# 	GAMAL = (GB * DMD) / GCR
# 	GAMMA = GAMAL / ALPHA
# 	GAMM2 = GAMMA - 2
# 	DELM2 = DELTA - 2
# 	TDC   = TDKOK - 273.15
# 	PSAT  = 10**( (0.7859+0.03477*TDC) / (1+0.00412*TDC) )\
# 				* (1 + PMBOK * (4.5E-6 + 6E-10 * TDC*TDC) )

# 	if PMBOK > 0:
# 		PWO = RHOK*PSAT/(1-(1-RHOK)*PSAT/PMBOK)
# 	else:
# 		PWO = 0

# 	W  = PWO*(1-DMW/DMD)*GAMMA/(DELTA-GAMMA)
# 	C1 = A*(PMBOK+W)/TDKOK
# 	C2 = (A*W+11.2684E-6*PWO)/TDKOK
	
# 	C3 = (GAMMA-1)*ALPHA*C1/TDKOK
# 	C4 = (DELTA-1)*ALPHA*C2/TDKOK
# 	C5 = 0
# 	C6 = 0

# 	# Conditions at the observer.
# 	R0  = S+HMOK
# 	TEMPO, DN0, RDNDR0 = sla_atmt(R0,TDKOK,ALPHA,GAMM2,DELM2,C1,C2,C3,C4,C5,C6,R0)
# 	SK0 = DN0*R0*np.sin(ZOBS2)
# 	F0  = sla_refi(DN0,RDNDR0)

# 	# Conditions in the troposphere at the tropopause.
# 	RT = S+max(HT,HMOK)
# 	TT, DNT, RDNDRT = sla_atmt(R0,TDKOK,ALPHA,GAMM2,DELM2,C1,C2,C3,C4,C5,C6,RT)
# 	SINE = SK0/(RT*DNT)
# 	ZT = np.arctan2(SINE,np.sqrt(max(1-SINE*SINE,0)))
# 	FT = sla_refi(DNT,RDNDRT)

# 	# Conditions in the stratosphere at the tropopause.
# 	DNTS,RDNDRP = sla_atms(RT,TT,DNT,GAMAL,RT)
# 	SINE = SK0/(RT*DNTS)
# 	ZTS  = np.arctan2(SINE,np.sqrt(max(1-SINE*SINE,0)))
# 	FTS  = sla_refi(DNTS,RDNDRP)

# 	# Conditions at the stratosphere limit.
# 	RS   = S+HS
# 	DNS,RDNDRS = sla_atms(RT,TT,DNT,GAMAL,RS)
# 	SINE = SK0/(RS*DNS)
# 	ZS   = np.arctan2(SINE,np.sqrt(max(1-SINE*SINE,0)))
# 	FS   = sla_refi(DNS,RDNDRS)

# 	# Variable initialization to avoid compiler warning.
# 	REFT = 0

# 	# Integrate the refraction integral in two parts;  first in the
# 	# troposphere (K=1), then in the stratosphere (K=2).
# 	for K in [1,2]:
# 		# Initialize previous refraction to ensure at least two iterations.
# 		REFOLD = 1

# 		# Start off with 8 strips.
# 		IS = 8

# 		# Start Z, Z range, and start and end values.
# 		if K==1:
# 			Z0 = ZOBS2
# 			ZRANGE = ZT-Z0
# 			FB = F0
# 			FF = FT
# 		else:
# 			Z0 = ZTS
# 			ZRANGE = ZS-Z0
# 			FB = FTS
# 			FF = FS

# 		# Sums of odd and even values.
# 		FO = 0
# 		FE = 0

# 		# First time through the loop we have to do every point.
# 		N = 1

# 		# Start of iteration loop (terminates at specified precision).
# 		LOOP = True
# 		while LOOP:

# 			# Strip width.
# 			H = ZRANGE/IS

# 			# Initialize distance from Earth centre for quadrature pass.
# 			if K==1:
# 				R = R0
# 			else:
# 				R = RT

# 			# One pass (no need to compute evens after first time).
# 			# DO I=1,IS-1,N
# 			for I in np.arange(1, IS-1, N):

# 				# Sine of observed zenith distance.
# 				SZ = np.sin(Z0+H*I)

# 				# Find R (to the nearest metre, maximum four iterations).
# 				if (SZ>1E-20):
# 					W  = SK0/SZ
# 					RG = R
# 					DR = 1E6
# 					J  = 0
# 					while (abs(DR)>1 and J<4):
# 						J+=1
# 						if K==1:
# 							TG,DN,RDNDR = sla_atmt(R0,TDKOK,ALPHA,GAMM2,DELM2,C1,C2,C3,C4,C5,C6,RG)
# 						else:
# 							DN,RDNDR = sla_atms(RT,TT,DNT,GAMAL,RG)
# 						DR = (RG*DN-W)/(DN+RDNDR)
# 						RG = RG-DR
# 					R = RG

# 				# Find the refractive index and integrand at R.
# 				if K==1:
# 					T,DN,RDNDR = sla_atmt(R0,TDKOK,ALPHA,GAMM2,DELM2,C1,C2,C3,C4,C5,C6,R)
# 				else:
# 					DN,RDNDR = sla_atms(RT,TT,DNT,GAMAL,R)

# 				F = sla_refi(DN,RDNDR)

# 				# Accumulate odd and (first time only) even values.
# 				if N==1 and (I%2)==0:
# 					FE = FE+F
# 				else:
# 					FO = FO+F


# 			# Evaluate the integrand using Simpson's Rule.
# 			REFP = H*(FB+4*FO+2*FE+FF)/3

# 			# Has the required precision been achieved (or can't be)?
# 			if (abs(REFP-REFOLD)>TOL and IS<ISMAX):
# 				# No: prepare for next iteration.

# 				# Save current value for convergence test.
# 				REFOLD = REFP

# 				# Double the number of strips.
# 				IS = IS+IS

# 				# Sum of all current values = sum of next pass's even values.
# 				FE = FE+FO

# 				# Prepare for new odd values.
# 				FO = 0

# 				# Skip even values next time.
# 				N = 2
# 			else:
# 				# Yes: save troposphere component and terminate the loop.
# 				if (K==1):
# 					REFT = REFP
# 				LOOP = False

# 	# Result.
# 	REF = REFT+REFP
# 	if (ZOBS1<0): 
# 		REF = -REF
# 	return REF






#!/usr/bin/env Python3
import numpy as np
from scipy import integrate
import os
from .ciddorModel import Observatory

'''
@ Author:		Joost van den Born 
@ Contact:		born@astron.nl
@ Description:	This script offers a complete model of the atmospheric refraction
				and dispersion. It uses the Ciddor1996 paper to calculate the 
				index of refraction for a wide range of conditions, with corresponding
				error propagation if needed.
				The actual refraction is calculated by any of the following models:
				- Plane parallel atmosphere
				- Cassini spherical homogeneous atmosphere
				- Mathar's Barometric exponential model
				- Other methods described in Corbard et al., 2019.
'''

class dispersion(Observatory):
	'''
		Class that sets the observer location.
		These location attributes are then used to calculate the 
		appropriate atmospheric dispersion.

		See also:
		Observatory()
		refraction()
		...

		Attributes
		-----------
		lat : 	float
			Latitude of the observer in degrees
		h 	: 	float
			Altitude above mean sea level of the observer in meters
		Other Observatory() attributes

		Methods
		-------
		cassini(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric dispersion assuming Cassini's homegeneous
			atmosphere model.
		cassiniError(z, l1, l2, T, p, RH, xc, dl1=0, dl2=0, dT=0.2, dP=20, 
					dRH=0.02, dCO2=20, dz=0, lat=None, h=None)
			Returns the uncertainty in the atmospheric dispersion, assuming 
			Cassini's homegeneous atmosphere model and the given parameters.
		corbard(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric dispersion following the method described 
			in Corbard et al., 2019.
		gammaCoefficients(n1, reduced_height=None)
			Helper function for matharExponential(), that calculates the gamma
			coefficients.
		H_isotherm(T)
			Sets the reduced height for the conditions at the observer, assuming
			an isothermal atmosphere.
		matharExponential(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric dispersion assuming	Mathar's barometric 
			exponential model.	
		oriani(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric dispersion following the classic version
			of Oriani's theorem.
		planeParallel(n1, n2, zenith)
			Returns the atmospheric dispersion assuming a plane parallel
			atmosphere.	
		psi(x)
			Submethod for corbard().
		refractionIntegral(n1, n2, zenith, R0=None, useStandardAtmosphere=True, heightData=None, rhoData=None)
			Returns the atmospheric dispersion following the refraction integral.	
		setReducedHeight(p, rho)
			Sets the reduced atmospheric height.
		set_H(p, rho)
			Sets the reduced height for the conditions at the observer, assuming
			an adiabatic atmosphere.
		set_rc()
			Sets the local radius of curvature of the Earth at the observer.
		tan5(n1, n2, zenith, reduced_height=None)
			Returns the atmospheric dispersion following Oriani's theorem,
			including a fifth order term in the expansion.		
		Other Observatory() methods
	'''
	def __init__(self, lat, h):
		Observatory.__init__(self)
		self.h  	= h 						# height above sea level in meters
		self.lat 	= lat						# Latitude of the observer in degrees
		self.rc 	= self.set_rc() * 1000		# Radius of curvature of the earth at the observer


	def cassini(self, n1, n2, zenith, reduced_height=None):
		'''
			Refraction of spherical atmosphere, derived from geometry using Sine law

			Parameters
			----------
			n1 : float
				Refractive index at the shortest wavelength of interest
			n2 : float
				Refractive index at the longest wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			dR : float
				Atmospheric dispersion in degrees
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H
		
		_R1 = np.arcsin(n1 * self.rc * np.sin(np.deg2rad(zenith)) / (self.rc + H)) - np.arcsin(self.rc * np.sin(np.deg2rad(zenith)) / (self.rc + H) )
		_R2 = np.arcsin(n2 * self.rc * np.sin(np.deg2rad(zenith)) / (self.rc + H)) - np.arcsin(self.rc * np.sin(np.deg2rad(zenith)) / (self.rc + H) )	
		return np.degrees(_R1) - np.degrees(_R2)


	def cassiniError(self, zenith, l1, l2, T, p, RH, xc, dl1=0, dl2=0, dT=0.2, dP=20,
						dRH=0.02, dCO2=20, dz=0, lat=None, h=None):
		'''
			Calculates the uncertainty in the atmospheric dispersion, 
			following the Cassini model, by using fully propagated error analysis.
			This involves almost endless use of the chain rule and evaluation 
			of both the moist air and dry air components. 
			Note: dry air does not contain any water, i.e. H=0, x_w=0, etc.

			Not taken into account is the errror propagation in the scale height
			of the atmosphere, from my results this should give a small or even 
			negligible difference.

			Parameters
			----------
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			l1 	: float
				Wavelength of light in microns, at the shortest wavelength 
				of interest
			l2 	: float
				Wavelength of light in microns, at the longest wavelength 
				of interest
			T 	: float
				Atmospheric temperature in Kelvin
			p 	: float
				Atmospheric pressure in Pascal
			RH 	: float
				Atmospheric relative humidity (0<=RH<=1)
			xc 	: float
				CO2 density in parts per million
			dl1	: float 
				Uncertainty in the wavelength in microns, at the shortest
				wavelength of interest
			dl2	: float 
				Uncertainty in the wavelength in microns, at the longest
				wavelength of interest
			dT 	: float (default: 0.2)
				Uncertainty in the atmospheric temperature in Kelvin
			dP 	: float (default: 20)
				Uncertainty in the atmospheric pressure in Pascal
			dRH : float (default: 0.02)
				Uncertainty in the atmospheric relative humidity
			dCO2: float (default: 20)
				Uncertainty in the atmospheric CO2 density in parts per million
			dz 	: float (default: 0)
				Uncertainty in the zenith angle, in degrees.
			lat : float (default: None)
				If entered, overrides the latitude attribute of the class.
			h 	: float (default=None)
				If entered, overrides the altitude attribute of the class.

			Returns
			-------
			dRR	: float
				Uncertainty in the atmospheric dispersion in degrees.
		'''
		if lat == None:
			lat = self.lat
		else:
			self.lat = lat
		if h == None:
			h 	= self.h
		else:
			self.h = h

		# constants
		_cf = 1.022
		_w0 = 295.235 	# um^-2
		_w1 = 2.6422	# um^-2
		_w2 = -0.032380	# um^-4
		_w3 = 0.004028	# um^-6

		_k0 = 238.0185 		# um^-2
		_k1 = 5792105  		# um^-2
		_k2 = 57.362		# um^-2
		_k3 = 167917		# um^-2	

		_A =  1.2378847e-5 # K^-2
		_B = -1.9121316e-2 # K^-2
		_C = 33.93711047
		_D = -6.3431645e3  # K^-2

		_Mw  = 0.018015   # kg/mol
		_R   = 8.314510	  # J mol^-1 K^-1, gas constant

		_a0 =  1.58123e-6 # K Pa^-1
		_a1 = -2.9331e-8  # Pa^-1
		_a2 =  1.1043e-10 # K^-1 Pa^-1
		_b0 =  5.707e-6   # K Pa^-1
		_b1 = -2.051e-8   # Pa^-1
		_c0 =  1.9898e-4  # K Pa^-1
		_c1 = -2.376e-6   # Pa^-1
		_d  =  1.83e-11   # K^2 Pa^-2
		_e  = -0.765e-8   # K^2 Pa^-2
		_t  = T - 273.15

		_aa = 1.00062
		_bb = 3.14e-8 	# Pa^-1
		_cc = 5.6e-7	# deg C^-2

		_RH_a = 0
		_RH_w = RH

		# Calculated values
		_rho_obs = self.rho(p, T, RH, xc)
		self.rc  = self.set_rc() * 1000
		self.H   = self.set_H(p, _rho_obs) 
		_alpha   = self.rc * np.sin(np.deg2rad(zenith)) / (self.rc + self.H)
		_p_a 	 = p - RH*self.svp(T) 
		_p_w     = RH*self.svp(T)
		_x_w_a	 = self.xw(_p_a, T, _RH_a)
		_x_w_w 	 = self.f(_p_w, T)
		_Z_a 	 = self.Z(p=_p_a, T=T, RH=_RH_a)
		_Z_w 	 = self.Z(p=_p_w, T=T, RH=_RH_w, xw=_x_w_w)
		_Ma  	 = 1e-3 * (28.9635 + 12.011e-6 * (xc - 400) )	# Molar mass in kg/mol of dry air containing xc ppm of CO2
		_p_sv 	 = self.svp(T)
		_rho_a 	 = self.rho(p=p-RH*self.svp(T), T=T, RH=0, xc=xc)
		_rho_axs = self.rho(p=101325, T=288.15, RH=0, xc=xc)
		if RH == 0.0:
			# nan is returned for _rho_w if we put in a (partial) pressure of 0
			_rho_w = 0
		else:
			_rho_w = self.rho(p=RH*self.svp(T), T=T, RH=RH, xc=xc) 
		_rho_ws  = self.rho(p=1333, T=293.15, RH=1333/self.svp(293.15), xc=xc)
		_n_a1 	 = self.n_axs(l1, xc)
		_n_w1 	 = self.n_ws(l1)
		_n_a2 	 = self.n_axs(l2, xc)
		_n_w2 	 = self.n_ws(l2)

		_L1_a 	 = (_n_a1**2 - 1) / (_n_a1**2 + 2)  # Dry air component of L
		_L1_w 	 = (_n_w1**2 - 1) / (_n_w1**2 + 2)
		_L1   	 = ( _rho_a / _rho_axs ) * _L1_a + ( _rho_w / _rho_ws ) * _L1_w # Overall L
		_L2_a 	 = (_n_a2**2 - 1) / (_n_a2**2 + 2) # Dry air component of L
		_L2_w 	 = (_n_w2**2 - 1) / (_n_w2**2 + 2)
		_L2    	 = ( _rho_a / _rho_axs ) * _L2_a + ( _rho_w / _rho_ws ) * _L2_w # Overall L

		_n1   	 = np.sqrt( (1 + 2 * _L1) / (1 - _L1) )
		_n2   	 = np.sqrt( (1 + 2 * _L2) / (1 - _L2) )

		# Partial derivatives for derivation to l1
		_dn_wdl1 = _cf * 1e-8 * (-1 * _w1 * 2/l1**3 - _w2 * 4/l1**5 - _w3 * 6/l1**7)
		_dLdn_w1 = _rho_w / _rho_ws * 6 * _n_w1 / (_n_w1**2 + 2)**2
		_dn_adl1 = 1e-8 * (1 + 0.534e-6 * (xc - 450)) * ( - (2 * _k1 * l1 ) / ((_k0 * l1**2 - 1)**2) - (2 * _k3 * l1) / ((_k2 * l1**2 - 1)**2) )
		_dLdn_a1 = _rho_a / _rho_axs * 6 * _n_a1 / (_n_a1**2 + 2)**2
		_dndL1   = ( ( (1 + 2*_L1) / (1 - _L1)**2 ) + (2 / (1-_L1)) ) / (2 * np.sqrt( (1 + 2*_L1) / (1 - _L1) ) ) 
		_dddn1   = _alpha / np.sqrt( 1 - _alpha**2 * _n1**2 )
		_dddl1   = _dddn1 * _dndL1 * ( _dLdn_a1 * _dn_adl1 + _dLdn_w1 * _dn_wdl1)

		# Partial derivatives for derivation to l2
		_dn_wdl2 = _cf * 1e-8 * (-1 * _w1 * 2/l2**3 - _w2 * 4/l2**5 - _w3 * 6/l2**7)
		_dLdn_w2 = _rho_w / _rho_ws * 6 * _n_w2 / (_n_w2**2 + 2)**2
		_dn_adl2 = 1e-8 * (1 + 0.534e-6 * (xc - 450)) * ( - (2 * _k1 * l2 ) / ((_k0 * l2**2 - 1)**2) - (2 * _k3 * l2) / ((_k2 * l2**2 - 1)**2) )
		_dLdn_a2 = _rho_a / _rho_axs * 6 * _n_a2 / (_n_a2**2 + 2)**2
		_dndL2   = ( ( (1 + 2*_L2) / (1 - _L2)**2 ) + (2 / (1-_L2)) ) / (2 * np.sqrt( (1 + 2*_L2) / (1 - _L2) ) ) 
		_dddn2   = _alpha / np.sqrt( 1 - _alpha**2 * _n2**2 )
		_dddl2   = _dddn2 * _dndL2 * ( _dLdn_a2 * _dn_adl2 + _dLdn_w2 * _dn_wdl2)


		# Partial derivatives for derivation to T
		# Dry air parts:
		_drho_adp_a  = _Ma / (_Z_a * _R * T) * (1 - _x_w_a * (1 - _Mw / _Ma))
		_dp_adp_sv 	 = - RH
		_dp_svdT 	 = (2*_A*T + _B - _D/(T**2)) * np.exp(_A * T**2 + _B * T + _C + _D/T )
		_drho_adTZ	 = - _p_a * _Ma / (_R * ( _Z_a * T)**2) * (1 - _x_w_a * (1 - _Mw / _Ma))
		_dTZdT_a	 = _p_a / T * (_a2 * _t**2 + _a1 * _t + _a0 + (_b1*_t + _b0) * _x_w_a + (_c0 + _c1*_t) * _x_w_a**2 ) - _p_a * (_a1 + 2*_a2*_t + _b1*_x_w_a + _c1*_x_w_a**2) - 2*(_p_a/T)**2 * (_d + _e * _x_w_a**2) + _Z_a
		_dTZdxw_a	 = 2*_e*_p_a**2 / T * _x_w_a - _p_a * (_b1*_t + _b0 + 2 * (_c0 + _c1*_t) * _x_w_a)
		_dxw_adf	 = _RH_a * _p_sv / _p_a
		_dfdT 		 = 2*_cc*_t
		_dxw_adp_sv	 = self.f(_p_a, T) * _RH_a / _p_a
		_drho_adxw_a = 0 # as rho_a has not water content, i.e. rho = p*M/ZRT


		# Moist air parts:
		_drho_wdp_w  = _Ma / (_Z_w * _R * T) * (1 - _x_w_w * (1 - _Mw / _Ma))
		_dp_wdp_sv 	 = RH
		_dp_svdT 	 = (2*_A*T + _B - _D/(T**2)) * np.exp(_A * T**2 + _B * T + _C + _D/T )
		_drho_wdTZ	 = - _p_w * _Ma / (_R * (_Z_w * T)**2) * (1 - _x_w_w * (1 - _Mw / _Ma))
		_dTZdT_w	 = _p_w / T * (_a2 * _t**2 + _a1 * _t + _a0 + (_b1*_t + _b0) * _x_w_w + (_c0 + _c1*_t) * _x_w_w**2 ) - _p_w * (_a1 + 2*_a2*_t + _b1*_x_w_w + _c1*_x_w_w**2) - 2*(_p_w/T)**2 * (_d + _e * _x_w_w**2) + _Z_w
		_dTZdxw_w	 = 2*_e*_p_w**2 / T * _x_w_w - _p_w * (_b1*_t + _b0 + 2 * (_c0 + _c1*_t) * _x_w_w)
		_dxw_wdf	 = 1 # _RH_w * _p_sv / _p_w
		_dfdT 		 = 2*_cc*_t
		_dxw_wdp_sv	 = self.f(_p_w, T) / _p_sv # * _RH_w / _p_w
		_drho_wdxw_w = _p_w * _Ma / (_Z_w * _R * T) * (_Mw / _Ma - 1)

		_tA =  _drho_adp_a * _dp_adp_sv * _dp_svdT +  _drho_adTZ * (_dTZdT_a + _dTZdxw_a * (_dxw_adf * _dfdT + _dxw_adp_sv * _dp_svdT) ) + _drho_adxw_a * ( _dxw_adf * _dfdT + _dxw_adp_sv * _dp_svdT )
		_tB =  _drho_wdp_w * _dp_wdp_sv * _dp_svdT +  _drho_wdTZ * (_dTZdT_w + _dTZdxw_w * (_dxw_wdf * _dfdT + _dxw_wdp_sv * _dp_svdT) ) + _drho_wdxw_w * ( _dxw_wdf * _dfdT + _dxw_wdp_sv * _dp_svdT )
		_dL1drho_a 	 = _L1_a / _rho_axs
		_dL2drho_a 	 = _L2_a / _rho_axs
		_dL1drho_w 	 = _L1_w / _rho_ws
		_dL2drho_w 	 = _L2_w / _rho_ws
		
		_dddT = _tA * (_dddn1 * _dndL1 * _dL1drho_a - _dddn2 * _dndL2 * _dL2drho_a) + _tB * (_dddn1 * _dndL1 * _dL1drho_w - _dddn2 * _dndL2 * _dL2drho_w)

		# Partial derivatives for derivation to p
		# Dry air parts:
		_drho_adZ 	= - _p_a * _Ma / ( _Z_a**2 * _R * T) * (1 - _x_w_a * (1 - _Mw / _Ma))
		_dZdp_a 	= - 1 / T * (_a0 + _a1 * _t + _a2 * _t**2 + (_b0 + _b1*_t)*_x_w_a + (_c0 + _c1*_t)*_x_w_a**2 ) + 2*_p_a/(T**2) * (_d + _e * _x_w_a**2)
		_dfdp 		= _bb
		_dxw_adp_a 	= - self.f(_p_a, T) * _RH_a * _p_sv / (_p_a)**2 

		# Moist air parts
		_dZdp_w 	= - 1 / T * (_a0 + _a1 * _t + _a2 * _t**2 + (_b0 + _b1*_t)*_x_w_w + (_c0 + _c1*_t)*_x_w_w**2 ) + 2*_p_w/(T**2) * (_d + _e * _x_w_w**2)
		# Not necessary, since dpwdp=0

		_pA = _drho_adp_a + _drho_adZ * _dZdp_a + _drho_adxw_a * (_dxw_adf * _dfdp + _dxw_adp_a)
		_dddP = _pA * (_dddn1 * _dndL1 * _dL1drho_a - _dddn2 * _dndL2 * _dL2drho_a)

		
		# Partial derivatives for derivation to H
		# Dry air parts:
		_dp_adH 	= - _p_sv
		_dZdxw_a 	= (_p_a / T)**2 * 2 * _e * _x_w_a - (_p_a / T) * ( (_b0 + _b1 * _t) + 2 * (_c0 + _c1 * _t) * _x_w_a )
		_dxw_adH 	= self.f(_p_a, T) * _p_sv / _p_a

		# Moist air parts:
		_dp_wdH 	= _p_sv
		_dZdxw_w 	= (_p_w / T)**2 * 2 * _e * _x_w_w - (_p_w / T) * ( (_b0 + _b1 * _t) + 2 * (_c0 + _c1 * _t) * _x_w_w )
		_dxw_wdH 	= 0
		_dxw_wdp_w 	= 0		
		_drho_wdZ 	= - _p_w * _Ma / ( _Z_w**2 * _R * T) * (1 - _x_w_w * (1 - _Mw / _Ma))

		_hA = _drho_adp_a * _dp_adH + _drho_adZ * ( _dZdp_a * _dp_adH + _dZdxw_a * (_dxw_adH + _dxw_adp_a * _dp_adH + _dxw_adf * _dfdp * _dp_adH) ) + _drho_adxw_a * (_dxw_adH + _dxw_adp_a * _dp_adH + _dxw_adf * _dfdp * _dp_adH)
		_hB = _drho_wdp_w * _dp_wdH + _drho_wdZ * ( _dZdp_w * _dp_wdH + _dZdxw_w * ( _bb * _p_sv ) ) + _drho_wdxw_w * ( _bb * _p_sv )
		_dddRH = _hA * (_dddn1 * _dndL1 * _dL1drho_a - _dddn2 * _dndL2 * _dL2drho_a) + _hB * (_dddn1 * _dndL1 * _dL1drho_w - _dddn2 * _dndL2 * _dL2drho_w)


		# Partial derivatives for derivation to xc
		# Dry air parts:
		_drho_adMa   = _p_a * (1 - _x_w_a) / (_Z_a * _R * T)
		_dL1drho_axs = - (_rho_a * _L1_a)/(_rho_axs**2)
		_dL2drho_axs = - (_rho_a * _L2_a)/(_rho_axs**2)
		_drho_axsdMa = _drho_adMa
		_dL1dL_a	 = _rho_a / _rho_axs
		_dL_adn_a1	 = 6 * _n_a1 / (_n_a1**2 + 2)**2
		_dL2dL_a	 = _dL1dL_a
		_dL_adn_a2	 = 6 * _n_a2 / (_n_a2**2 + 2)**2

		# Moist air parts:
		_drho_wdMa   = _p_w * (1 - _x_w_w) / (_Z_w * _R * T)
		_dL1drho_ws = - (_rho_w * _L1_w)/(_rho_ws**2)
		_dL2drho_ws = - (_rho_w * _L2_w)/(_rho_ws**2)
		_drho_wsdMa = _drho_wdMa

		# Other parts:
		_dMadx_c   = 12.011e-9
		_dn_adx_c1 = (self.n_as(l1) - 1) * 0.534e-6
		_dn_adx_c2 = (self.n_as(l2) - 1) * 0.534e-6

		_dL1dx_c = _dL1drho_a * _drho_adMa * _dMadx_c + _dL1drho_w * _drho_wdMa * _dMadx_c + _dL1drho_axs * _drho_axsdMa * _dMadx_c + _dL1drho_ws * _drho_wsdMa * _dMadx_c + _dL1dL_a * _dL_adn_a1 * _dn_adx_c1
		_dL2dx_c = _dL2drho_a * _drho_adMa * _dMadx_c + _dL2drho_w * _drho_wdMa * _dMadx_c + _dL2drho_axs * _drho_axsdMa * _dMadx_c + _dL2drho_ws * _drho_wsdMa * _dMadx_c + _dL2dL_a * _dL_adn_a2 * _dn_adx_c2
		_dddCO2 = _dddn1 * _dndL1 * _dL1dx_c - _dddn2 * _dndL2 * _dL2dx_c


		# Partial derivatives for derivation to z
		_zenith = np.deg2rad(zenith)
		_dz   	= np.deg2rad(dz)
		_R 		= self.rc
		_H 		= self.H
		_dddz 	= (_R * np.cos(_zenith)) / (_R + _H) * ( _n1 / np.sqrt(1-(_n1**2 * _R**2 * np.sin(_zenith)**2)/(_R + _H)**2) - _n2 / np.sqrt(1-(_n2**2 * _R**2 * np.sin(_zenith)**2)/(_R + _H)**2) )		

		# Final error:
		_dispersionError = np.sqrt( (_dddl1 * dl1)**2 + (_dddl2 * dl2)**2 + (_dddT * dT)**2  + (_dddP * dP)**2 + (_dddRH * dRH)**2 + (_dddCO2 * dCO2)**2  + (_dddz * _dz)**2)

		return np.degrees(_dispersionError)


	def corbard(self, n1, n2, zenith, reduced_height=None):
		'''
			Corbard et al., 2019, mention an additional formula based on
			the error function. Oriani's theorem can be derived from this
			equation by 'keeping only the three first terms of its 
			asymptotic expansion'.

			Parameters
			----------
			n1 : float
				Refractive index at the shortest wavelength of interest
			n2 : float
				Refractive index at the longest wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			dR : float
				Atmospheric dispersion in degrees
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H

		_rList = []
		for _n in [n1, n2]:
			_a = _n - 1
			_b = H / self.rc
			_R = _a * ( (2 - _a) / (np.sqrt( 2*_b - _a )) ) * np.sin( np.deg2rad(zenith) ) * self.psi( (np.cos(np.deg2rad(zenith))) / np.sqrt(2*_b - _a) )
			_rList.append(np.degrees(_R))
		return _rList[0] - _rList[1]


	def gammaCoefficients(self, n1, reduced_height=None):
		'''
			Submethod for matharExponential(). Calculates the gamma
			coefficients.

			Parameters
			----------
			n1 	: 	float
				Refractive index at the wavelength of interest at the observer
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			g 	: 	array
				List of values, corresponding to the gamma coefficients.
				g[0] corresponds to g_1, g[1] to g_3, g[2] to g_5, and so on.
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H

		_chi0   = n1**2 - 1
		_Khat 	= H / self.rc

		_g1 = (1 - 5/24 * _Khat * _chi0**2 - 3/4 * _chi0 - 3/8 * _Khat**2 * _chi0 \
			- _Khat - 6 * _Khat**3 + 5/8 * _chi0**2 + 63/128 * _chi0**4 + 3/8 \
			* _Khat * _chi0 + 35/256 * _Khat * _chi0**3 + 2 * _Khat**2 + 5/36 \
			* _Khat**2 * _chi0**2 - 35/64 * _chi0**3 + 9/16 * _Khat**3 * _chi0 \
			- 120 * _Khat**5 + 24 * _Khat**4 - 231 / 512 * _chi0**5 - 63 / 640 \
			* _Khat * _chi0**4 - 35 / 512 * _Khat**2 * _chi0**3 - 5 / 36 * _Khat**3\
			* _chi0**2 - 9 / 8 * _Khat**4 * _chi0 )
		
		_g3 = (35/96 * _chi0**3 - 5/12 * _chi0**2 + 333 * _Khat**4 * _chi0 - 2400 \
			* _Khat**5 + 3605/4608 * _Khat**2 * _chi0**3 + 63/256 * _Khat * _chi0**4\
			 - 819/16 * _Khat**3 * _chi0 - 35/18 * _Khat**2 * _chi0**2 - 35/96 \
			 * _Khat * _chi0**3 + 69/8 * _Khat**2 * _chi0 + 5/8 * _Khat * _chi0**2 \
			 - 3/2 * _Khat * _chi0 + 1355/216 * _Khat**3 * _chi0**2 + 10 * _Khat**2 \
			 + 336 * _Khat**4 - 2 * _Khat + _chi0/2 - 21/64 * _chi0**4 - 54 * _Khat**3 \
			 + 77/256 * _chi0**5 ) * 1/2
		
		_g5 = (-96 * _Khat**3 + 8 * _Khat**2 - 7/24 * _chi0**3 + 65/4 * _Khat**2 \
			* _chi0**2 + 35/48 * _Khat * _chi0**3 + 30 * _Khat**2 * _chi0 - 5/3 \
			* _Khat * _chi0**2 - 3 * _Khat * _chi0 + 20091/8 * _Khat**4 * _chi0 \
			- 32605/216 * _Khat**3 * _chi0**2 - 35/9 * _Khat**2 * _chi0**3 + 984 \
			* _Khat**4 + _chi0**2/3 - 10200 * _Khat**5 - 77/320 * _chi0**5 + 21/80 \
			* _chi0**4 - 1089/4 * _Khat**3 * _chi0 - 7/16 * _Khat * _chi0**4 ) * 3/8
		
		_g7 = (-17520 * _Khat**5 - 9/40 * _chi0**4 - 8755/12 * _Khat**3 * _chi0**2 \
			+ 1056 * _Khat**4 + _chi0**3/4 + 23517/4 * _Khat**4 * _chi0 - 11/3 \
			* _Khat * _chi0**2 + 21 * _Khat**2 * _chi0 - 7/4 * _Khat * _chi0**3 \
			+ 505/9 * _Khat**2 * _chi0**2 - 801/2 * _Khat**3 * _chi0 + 33/160 \
			* _chi0**5 - 48 * _Khat**3 + 63/80 * _Khat * _chi0**4 + 2345/96 \
			* _Khat**2 * _chi0**3) * 5/16
		
		_g9 = (_chi0**4 / 5 + 384 * _Khat**4 - 13440 * _Khat**5 + 5562 * _Khat**4 \
			* _chi0 + 3125/36 * _Khat**2 * _chi0**3 - 11/60 * _chi0**5 - 25/6 \
			* _Khat * _chi0**3 - 26950/27 * _Khat**3 * _chi0**2 + 340/9 * _Khat**2 \
			* _chi0**2 - 180 * _Khat**3 * _chi0 - 9/5 * _Khat * _chi0**4) * 35/128

		g = np.array([_g1, _g3, _g5, _g7, _g9])
		g*= n1*_chi0/2

		return g


	def H_isotherm(self, T):
		'''
			Calculates the reduced height of the atmosphere, 
			assuming an isotherm distribution.

			Parameters
			----------
			T 	: float
				Atmospheric temperature in Kelvin

			Returns
			-------
			H 	: float
				Reduced height of the atmosphere in meters
		'''
		_kb = 1.38064852e-23 # kg m2 s-2 K-1
		_m  = 4.809651698e-26 # kg (weight per molecule of air, assuming dry air)

		_phi = np.deg2rad(self.lat)
		_c1 = 5.2790414e-3
		_c2 = 2.32718e-5
		_c3 = 1.262e-7
		_c4 = 7e-10
		_g0 = 9.780327 # m/s^2
		_g0_local = _g0 * (1 + _c1 * np.sin(_phi)**2 + _c2 * np.sin(_phi)**4 + _c3 * np.sin(_phi)**6 + _c4 * np.sin(_phi)**8)

		return (_kb*T) / (_m * _g0_local)


	def matharExponential(self, n1, n2, zenith, reduced_height=None):
		'''
			Model on atmospheric refraction described by Richard Mathar in 
			"A Barometric Exponential Model of the Atmosphere’s Refractive Index:  ZenithAngles and Second Order Aberration in the Entrance Pupil" (2020)
			ArXiv id: 2004.11808 (https://arxiv.org/abs/2004.11808)
			
			The code in this function is adapted from eqs 3, 16, 17 and 18 in this
			work. It is an adaptation of the RefrExpo.cxx C++ file written by Mathar.

			This model assumes the refractive index, n, is related to the susceptibility, 
			chi, following: n(r) = sqrt(1+chi(r)) (eq. 9).

			R. Mathar then assumes that the susceptibility is proportional to the
			air density and reduced height, following eq. 10:
			chi(r) = chi0 * exp(-(r-rho)/K).

			The refraction is described using a Taylor expansion; R in terms of
			the odd powers of the tangent of the observed zenith angle (eq. 3), not
			unlike Oriani's theorem. This requires coefficients g_i.

			Parameters
			----------
			n1 : float
				Refractive index at the wavelength of interest at the observer
			n2 : float
				Refractive index at the longest wavelength of interest at the
				observer
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			dR : float
				Atmospheric refraction in degrees
		'''
		if reduced_height:
			_H = reduced_height
		else:
			_H = self.H

		_zenith	= np.deg2rad(zenith)
		_rc 	= self.rc # rho

		# Atmospheric refraction for the first wavelength
		_gamma1 = self.gammaCoefficients(n1, reduced_height=_H)
		_R1 = sum([_gamma1[i]*np.tan(_zenith)**(2*i+1) for i in range(len(_gamma1))])

		# Atmospheric refraction for the second wavelength
		_gamma2 = self.gammaCoefficients(n2, reduced_height=_H)
		_R2 = sum([_gamma2[i]*np.tan(_zenith)**(2*i+1) for i in range(len(_gamma2))])

		# The atmospheric dispersion in radians
		_dR = _R1 - _R2

		return np.degrees(_dR)


	def oriani(self, n1, n2, zenith, reduced_height=None):
		'''
			Classic refraction formula with two tan(z) terms. 
			This does not assume any information about the structure
			of the atmosphere. Found with Oriani's theorem.

			Parameters
			----------
			n1 : float
				Refractive index at the shortest wavelength of interest
			n2 : float
				Refractive index at the longest wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			dR : float
				Atmospheric dispersion in degrees
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H

		_rList = []
		for _n in [n1, n2]:
			_a = _n - 1
			_b = H / self.rc
			_R = _a * (1 - _b) * np.tan( np.deg2rad(zenith) ) - _a * (_b - _a/2) * np.tan( np.deg2rad(zenith) )**3
			_rList.append(np.degrees(_R))
		return _rList[0] - _rList[1]


	def planeParallel(self, n1, n2, zenith):
		'''
			Refraction based on a flat atmosphere, easily derived from Snell's law.

			Parameters
			----------
			n1 : float
				Refractive index at the shortest wavelength of interest
			n2 : float
				Refractive index at the longest wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.

			Returns
			-------
			dR : float
				Atmospheric dispersion in degrees
		'''
		_R1 = np.arcsin(n1 * np.sin( np.deg2rad(zenith) )) - np.deg2rad(zenith)
		_R2 = np.arcsin(n2 * np.sin( np.deg2rad(zenith) )) - np.deg2rad(zenith)
		return np.degrees(_R1) - np.degrees(_R2)


	def psi(self, x):
		'''
			Submethod for corbard(). Described explicitly in Corbard et al.,2019.

			Parameters
			----------
			x 	: 	float
				Input of the function

			Returns
			-------
			y 	: 	float
				Output of the function
		'''
		_f   = lambda a: np.exp(-a**2)
		_int = integrate.quad(_f, x, np.inf)
		return np.exp(x**2) * _int[0]


	def refractionIntegral(self, n1, n2, zenith, R0=None, useStandardAtmosphere=True, 
				heightData=None, rhoData=None):
		'''
			Using data that gives temperature, pressure and density as a 
			function of height, the gladstone-dale relation is invoked to
			calculate the path of a light ray along the atmosphere. 
			This is allows us to also include atmospheric activity in 
			different atmosphere layers.
			Recommended use with the US Standard Atmosphere from 1976.
			USSA1976 data generated from code supplied by http://www.pdas.com/atmos.html
			The method described in eq. 3 in Auer & Standish (2000) is used for the 
			calculation.

			Parameters
			----------
			n1 : float
				Refractive index at the shortest wavelength of interest
			n2 : float
				Refractive index at the longest wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			R0 : float (optional)
				Local radius of curvature of the Earth in km.
			useStandardAtmosphere : bool
				Flag to use the included USSA1976 data
			heightData : 1D Array
				If "useStandardAtmosphere" is False, the function will ask
				for alternative atmospheric data to integrate over. This
				gives the altitudinal points
			rhoData : 1D Array
				If useStandardAtmosphere is False, the function will ask
				for alternative atmospheric data to integrate over. This
				gives the atmospheric density at the points defined by 
				"heightData".

			Returns
			-------
			dR : float
				Atmospheric dispersion in degrees
		'''
		if not R0:
			R0 = self.rc
		if useStandardAtmosphere:
			_datafileLocation = os.path.join(os.path.dirname(__file__), 'data/1976USSA.txt')
			_h, _, _, _, _Temp, _Press, _Dens,_,_,_ = np.genfromtxt(_datafileLocation, unpack=True, skip_header=6)
		else:
			_Dens = rhoData
			_h 	  = heightData 

		
		_rList 	= []
		zenith 	= np.deg2rad(zenith)
		
		# Integration of the refraction integral over the atmosphere layers.
		for _n in [n1, n2]:
			_Dens 	= np.asarray(_Dens)
			_h 	 	= np.asarray(_h) 
			_R 		= R0 + _h * 1000
			
			# Gladstone-Dale relation between refractive index and density
			_n_h = 1 + (_n - 1) / _Dens[0] * _Dens 

			# Calculation of the zenith angle for given height and refractive index, from the refractive invariant
			_z 		= np.arcsin( (_n * R0) / (_n_h * _R) * np.sin(zenith) )
			
			# Calculation of log space derivative of n and r
			_lnn = np.log(_n_h)
			_lnr = np.log(_R)
			_dlnn = np.asarray([_lnn[i+1] - _lnn[i] for i in range(len(_lnn)-1)])
			_dlnr = np.asarray([_lnr[i+1] - _lnr[i] for i in range(len(_lnr)-1)])
			_dz   = np.asarray([(_z[i] - _z[i+1]) for i in range(len(_z)-1)])

			# Calculation of the integrand in eq. 3 in Auer & Standish
			_Integrand = -1*(_dlnn/_dlnr) / (1 + _dlnn/_dlnr) * _dz
			_rList.append(np.degrees(np.sum(_Integrand)))
		return _rList[0] - _rList[1]


	def setReducedHeight(self, p, rho):
		'''
			Sets the reduced atmospheric height.

			Parameters
			----------
			p 	: float
				Atmospheric pressure in Pascal
			rho : float
				Density of the atmospheric air at the observer
		'''
		self.H  	= self.set_H(p, rho) 		# Reduced height of the atmosphere assuming ideal gas law


	def set_H(self, p, rho):
		'''
			Calculates the reduced height of the atmosphere, assuming ideal gas law.

			Parameters
			----------
			p 	: float
				Atmospheric pressure in Pascal
			rho : float
				Density of the atmospheric air at the observer

			Returns
			-------
			H 	: float
				Reduced height of the atmosphere in meters
		'''
		_phi = np.deg2rad(self.lat)
		_c1 = 5.2790414e-3
		_c2 = 2.32718e-5
		_c3 = 1.262e-7
		_c4 = 7e-10
		_g0 = 9.780327 # m/s^2
		_g0_local = _g0 * (1 + _c1 * np.sin(_phi)**2 + _c2 * np.sin(_phi)**4 + _c3 * np.sin(_phi)**6 + _c4 * np.sin(_phi)**8)
		_g = _g0_local - (3.0877e-6 - 4.3e-9 * np.sin(_phi)**2) * self.h + 7.2e-13 * self.h**2
		return p / (rho * _g)


	def set_rc(self):
		'''
			Returns the radius of the curvature at the observer in km,
			taking into account the latitude of the observer and the corresponding
			ellipsoidality of the earth.

			Returns
			-------
			rc 	: float
				Local radius of curvature of the Earth in kilometers.
		'''
		_A   = 6378.137 # km
		_B   = 6356.752 # km
		_phi = np.deg2rad(self.lat)
		
		_rc0 = (_A*_B)**2 / (_A**2 * np.cos(_phi)**2 + _B**2 * np.sin(_phi)**2)**(3/2)
		return _rc0 + self.h/1000


	def tan5(self, n1, n2, zenith, reduced_height=None):
		'''
			Same as oriani's theorem, but including a higher order term.
			Found in Corbard et al., 2019.

			Parameters
			----------
			n1 : float
				Refractive index at the shortest wavelength of interest
			n2 : float
				Refractive index at the longest wavelength of interest
			zenith : float
				Angle of observation in degrees.
				zenith = 0 means the observation is directly overhead.
			reduced_height : float (optional)
				Reduced height of the atmosphere. Default value is calculated
				from the object attributes.

			Returns
			-------
			dR : float
				Atmospheric dispersion in degrees
		'''
		if reduced_height:
			H = reduced_height
		else:
			H = self.H
		
		_rList = []
		for _n in [n1, n2]:
			_a = _n - 1
			_b = H / self.rc
			_R = _a * (1 - _b) * np.tan( np.deg2rad(zenith) )  - _a * (_b - _a/2) * np.tan( np.deg2rad(zenith) )**3 + 3 * _a * (_b - _a/2)**2 * np.tan( np.deg2rad(zenith) )**5
			_rList.append(np.degrees(_R))
		return _rList[0] - _rList[1]



#!/usr/bin/env python3

'''
AstroAtmosphere: A python package to evaluate refractive indices, refraction
                 and dispersion of the atmosphere. It was written for astronomic 
                 applications, but it can probably be used for other applications
                 as well.
'''
from setuptools import setup


with open('README.md') as f:
    __readme__ = f.read()

with open('LICENSE.txt') as f:
    __license__ = f.read()


def setup_package():
    # Rewrite the version file every time
    # write_version_py()

    setup(name='AstroAtmosphere',
          version='1.4',
          description='Collection of equations for refractive index, refraction and dispersion calculations of atmospheric air.',
          long_description=__readme__,
          long_description_content_type='text/markdown',
          author='Joost van den Born',
          author_email='born@astron.nl',
          url='https://gitlab.astro-wise.org/micado/atmosphericmodel',
          download_url='https://gitlab.astro-wise.org/micado/atmosphericmodel/-/archive/v1.4/atmosphericmodel-v1.4.tar.gz',
          license='GNU General Public License v3',
          data_files = [('',['LICENSE.txt'])],
          package_dir={'AstroAtmosphere': 'AstroAtmosphere'},
          package_data={'AstroAtmosphere': ['data/*.txt']},
          include_package_data=True,
          packages=['AstroAtmosphere'],
          keywords=['REFRACTION','REFRACTIVITY','ATMOSPHERE','DISPERSION','CIDDOR'],
          install_requires=['numpy',
                            'scipy'],
          extras_require={'slalib':['pyslalib>=1.0.4']},
          classifiers=['Programming Language :: Python :: 3',
                       'Operating System :: OS Independent',
                       'Intended Audience :: Science/Research',
                       'Topic :: Scientific/Engineering :: Astronomy', ]
          )


if __name__ == '__main__':
    setup_package()